# -*- coding: utf-8 -*-

"""Module to handle pdb files."""

import os
import logger
import re
import gzip
from urllib2 import urlopen, HTTPError, URLError
from StringIO import StringIO
import json

_name = "PDB"

class Pdb:
    """
    Pdb parser. Initialized by:
    1. pdb filename
    2. gzipped pdb filename
    3. 4-letter pdb code
    """

    def __init__(self,*args, **kwargs):
        self.file_name = None
        self.pdb_code = None
        self.dir = os.getcwd()
        self.loc = os.path.join(self.dir, "input_pdb")
        self.codification = {"ALA" : 'A', "CYS" : 'C', "ASP" : 'D', "GLU" : 'E', "PHE" : 'F', "GLY" : 'G', "HIS" : 'H',
                             "ILE" : 'I', "LYS" : 'K', "LEU" : 'L', "MET" : 'M', "MSE" : 'M', "ASN" : 'N', "PYL" : 'O',
                             "PRO" : 'P', "GLN" : 'Q', "ARG" : 'R', "SER" : 'S', "THR" : 'T', "SEC" : 'U', "VAL" : 'V',
                             "TRP" : 'W', "5HP" : 'E', "ABA" : 'A', "AIB" : 'A', "BMT" : 'T', "CEA" : 'C', "CGU" : 'E',
                             "CME" : 'C', "CRO" : 'X', "CSD" : 'C', "CSO" : 'C', "CSS" : 'C', "CSW" : 'C', "CSX" : 'C',
                             "CXM" : 'M', "DAL" : 'A', "DAR" : 'R', "DCY" : 'C', "DGL" : 'E', "DGN" : 'Q', "DHI" : 'H',
                             "DIL" : 'I', "DIV" : 'V', "DLE" : 'L', "DLY" : 'K', "DPN" : 'F', "DPR" : 'P', "DSG" : 'N',
                             "DSN" : 'S', "DSP" : 'D', "DTH" : 'T', "DTR" : 'X', "DTY" : 'Y', "DVA" : 'V', "FME" : 'M',
                             "HYP" : 'P', "KCX" : 'K', "LLP" : 'K', "MLE" : 'L', "MVA" : 'V', "NLE" : 'L', "OCS" : 'C',
                             "ORN" : 'A', "PCA" : 'E', "PTR" : 'Y', "SAR" : 'G', "SEP" : 'S', "STY" : 'Y', "TPO" : 'T',
                             "TPQ" : 'F', "TYS" : 'Y', "TYR" : 'Y' }
        keys = self.codification.keys()
        self.sequences = {}
        self.onlycalfa = ""
        self.allatoms = ""
        self.chain = ""
        self.canumber = 0
        self.allnumber = 0

        if args and len(args) == 1:
            if args[0] is None: raise logger.AggrescanError("No pdb code/file provided. Quitting.",
                                                            module_name=_name)
            if os.path.isfile(args[0]):
                self.file_name = args[0]
            else:
                self.pdb_code = args[0]
        if kwargs:
            self.loc = kwargs['output']
            try:
                self.chain = kwargs['chain']
            except KeyError:
                pass

        if self.file_name:
            try:
                self.handler = gzip.GzipFile(filename=self.file_name)
                self.data = self.handler.readlines()
                logger.debug(module_name=_name, msg="Reading %s" % os.path.abspath(self.file_name))
            except IOError:
                try:
                    self.handler = open(self.file_name)
                    self.data = self.handler.readlines()
                    logger.debug(module_name=_name, msg="Reading %s" % os.path.abspath(self.file_name))
                except IOError:
                    raise logger.AggrescanError("Couldnt open specified filename %s. Quitting.' % os.path.abspath(self.file_name)",
                                                module_name=_name)
        elif self.pdb_code:
            self.handler = self.download_pdb()
            self.data = self.handler.readlines()

        seq = re.compile(r"^ATOM.{9}CA..(?P<seqid>.{3}).(?P<chain>.{1})(?P<resid>.{4})")  # TODO zle dla alternatywnych
        if self.chain != '':
            atm = re.compile(r"^ATOM.{9}(.{2}).( |A).{4}" + self.chain + "(?P<resid>.{4})(?P<x>.{12})(?P<y>.{8})(?P<z>.{8})")
        else:
            atm = re.compile(r"^ATOM.{9}(.{2}).( |A).{5}(?P<resid>.{4})(?P<x>.{12})(?P<y>.{8})(?P<z>.{8})")

        ter = re.compile(r'^END|^TER')
        mod = re.compile(r"^ENDMDL")
        self.trajectory = []
        self.sequence = ""

        lines = self.data
        end = len(lines) - 1
        counter = 0
        self._chainsOrder(lines)
        self._resIndexes(lines)
        self.mutatedata = {}

        for line in lines:
            line = re.sub(r'^HETATM(.{11})MSE(.*$)', r'ATOM  \1MET\2', line)
            localData = atm.match(line)
            data_seq = seq.match(line)

            if data_seq:
                seqid = data_seq.groups()[0].strip()
                chainid = data_seq.groups()[1].strip()
                resid = data_seq.groups()[2].strip()

                if seqid in keys:
                    s = self.codification[seqid]
                else:
                    s = "X"
                self.sequence += s

                # add to mutate page
                if chainid in self.mutatedata.keys():
                    self.mutatedata[chainid].append({'chain': chainid,
                                                     'resname': s,
                                                     'residx': resid})
                else:
                    self.mutatedata[chainid] = [{'chain': chainid,
                                                 'resname': s,
                                                 'residx': resid}]

                if chainid in self.sequences.keys():
                    self.sequences[chainid] += s
                else:
                    self.sequences[chainid] = s

            if localData:
                self.allnumber += 1
                self.allatoms += line
                dg = localData.groups()
                if dg[0] == 'CA':
                    self.onlycalfa += line
                    self.canumber += 1

            if counter == end:
                self.onlycalfa += line
                self.allatoms += line
            if ter.match(line):
                if self.chain:
                    if line[21] == self.chain:
                        self.onlycalfa += line
                        self.allatoms += line
                else:
                    self.onlycalfa += line
                    self.allatoms += line

            if (mod.match(line) and len(self.onlycalfa) > 1) or counter == end:
                break
            counter += 1
        self.handler.close()

    def _resIndexes(self, body):
        atm = re.compile(r"^ATOM.{9}CA..(?P<seqid>.{3}).(?P<chain>.{1})(?P<resid>.{4})")
        ter = re.compile(r'^END|^TER')
        mod = re.compile(r"^ENDMDL")
        self.numb = {}
        for chain in self.chains_order:
            self.numb[chain] = []

        for line in body:
            d = atm.match(line)
            if d:
                self.numb[d.group('chain').strip()].append(int(d.group('resid')))
            if mod.match(line):
                break

    def _chainsOrder(self, body):
        atm = re.compile(r"^ATOM.{9}CA..(?P<seqid>.{3}).(?P<chain>.{1})(?P<resid>.{4})")
        self.chains_order = []
        for line in body:
            d = atm.match(line)
            if d and d.group('chain') not in self.chains_order:
                self.chains_order.append(d.group('chain'))

    def isSingleChain(self):
        if self.chain != '' or len(self.sequences.keys()) == 1:
            return True
        else:
            return False

    def containsOnlyCA(self):
        if self.allnumber == self.canumber:
            return True
        else:
            return False

    def isBroken(self):
        brk = []
        if self.chain != '':
            indexes = self.numb[self.chain]
            first = indexes[0]
            for i in range(1, len(indexes)):
                if indexes[i] - 1 != first:
                    brk.append(str(first) + "-" + str(indexes[i]))
                first = indexes[i]
        else:
            for chain in self.sequences.keys():
                indexes = self.numb[chain]
                first = indexes[0]
                for i in range(1, len(indexes)):
                    if indexes[i] - 1 != first:
                        brk.append(str(first) + "-" + str(indexes[i]))
                    first = indexes[i]
        if len(brk) > 0:
            return ", ".join(brk)
        return False

    def getResIndexes(self):
        t = [str(i) for i in self.numb[self.chain]]
        return ",".join(t)

    def getBody(self):
        return self.allatoms

    def containsChain(self, chain):
        if chain in self.sequences.keys():
            return True

    def getSequenceNoHTML(self):
        if self.chain != '':
            return self.sequences[self.chain]
        else:
            out = ""
            for k in self.sequences.keys():
                out += "".join(self.sequences[k])
            return out

    def getSequence(self):
        if self.chain != '':
            return "<strong>" + self.chain + "</strong>: " + self.sequences[self.chain]
        else:
            out = ""
            for k in self.sequences.keys():
                out += "<strong>" + k + "</strong>: "
                out += "".join(self.sequences[k])
                out += "<br>"
            return out

    def getChainIdxResname(self):
        if self.chain == '':
            return json.dumps(self.mutatedata)
        else:
            return json.dumps({self.chain: self.mutatedata[self.chain]})

    def savePdbFile(self,path=''):
        if path:
            logger.to_file(filename=path, content=self.allatoms, allow_err=True)
        else:
            logger.to_file(filename=self.loc, content=self.allatoms, allow_err=True)

    def getPath(self):
        if os.path.isfile(self.loc):
            return self.loc
        else:
            raise logger.AggrescanError("Location for pdb file requested at: %s. The file was not found." % self.loc,
                                        module_name=_name)

    def download_pdb(self):
        try:
            gz_string = urlopen('http://www.rcsb.org/pdb/files/' + self.pdb_code.lower() + '.pdb.gz').read()
        except HTTPError as e:
            raise logger.AggrescanError("Could not download the pdb file. %s is not a valid pdb code/file. " % self.pdb_code,
                                        module_name=_name)

        except URLError as e:
            raise logger.AggrescanError("Could not download the pdb file. Can't connect to the PDB database - quitting",
                                        module_name=_name)
        fileLike = StringIO(gz_string)
        logger.debug(module_name=_name, msg="Successfully downloaded %s" % self.pdb_code.lower() + '.pdb.gz')
        return gzip.GzipFile(fileobj=fileLike,mode="rb")

    def validate(self):
        logger.debug(module_name=_name,msg='Validating pdb file: %s' % self.loc)
        if self.chain != '' and not self.containsChain(self.chain):
            raise logger.AggrescanError("Selected chain: %s not found in the pdb file. Quitting." % self.chain,
                                        module_name=_name)
        seq = self.getSequence()
        seq = re.sub("<strong>\w+</strong>:", "", seq)
        seq = re.sub("<br>", "", seq)
        seq = seq.replace(" ", "")
        allowed_seq = ['A', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'K', 'L',
                       'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V',
                       'W', 'Y']
        if len(seq) < 4:
            raise logger.AggrescanError("Sequence too short (perhaps something went wrong with pdb parsing).",
                                        module_name=_name)
        for e in seq:
            if e not in allowed_seq:
                raise logger.AggrescanError("Not supported amino acid: %s found in pdb file. Quitting." % e,
                                             module_name=_name)



if __name__ == '__main__':
    pass



# -*- coding: utf-8 -*-

from bokeh.plotting import figure, output_file, ColumnDataSource, save, show
from bokeh.models import HoverTool, FactorRange, Legend
from bokeh.models.glyphs import Line
from bokeh.resources import INLINE
from bokeh.util.string import encode_utf8
import bokeh.palettes
from bokeh.embed import components
from os.path import join

aa_dict_F = {'A': 'alanine', 'R': 'arginine', 'N': 'asparagine',
             'D': 'aspartic acid', 'C': 'cysteine', 'E': 'glutamic acid',
             'Q': 'glutamine', 'G': 'glycine', 'H': 'histidine',
             'I': 'isoleucine', 'L': 'leucine', 'K': 'lysine',
             'M': 'methionine', 'F': 'phenylalanine', 'P': 'proline',
             'S': 'serine', 'T': 'threonine', 'W': 'tryptophan',
             'Y': 'tyrosine', 'V': 'valine', 'X': 'unknown'}


def create_mut_plot(data_dir, save_plot=False):
    # This will read a mutant list that is already available to the calling jscript, dunno if its worth to pass it tho
    mutants = [""]      # The first is a fake - it will read the base simulation result and name it Base
    max_mutants, counter = 12, 0  # This is also used somewhere else so maybe should unify it
    p = []  # To create the figure object in a loop
    with open(join(data_dir, "Mutations_summary.csv"), "r") as f:
        f.readline()
        for line in f:
            mutants.append(line.split(",")[0])
            counter += 1
            if counter >= max_mutants:
                break

    counter = 0
    colors = bokeh.palettes.Category10[10]  # This is not enough colors need 13 for 12 mutants and its only 10
    colors.append('#1E1717')    # almost black
    colors.append('#f3ff00')    # strong, flashy yellow
    colors.append('#ce6778')    # kinda pinkish I guess
    legend_items = []
    for mutant in mutants:
        x, y, chain, name, index, status = [], [], [], [], [], []
        mutant = mutant if mutant else "A3D"    # It's a band aid considering file locations and names have changed
                                                # (Base csv is named A3D.csv)
        with open(join(data_dir, mutant + ".csv"), 'r') as f:
            if mutant == "A3D":     # Following the weird convention to get the Base case first with the name Base
                mutant = "Wild type"
            f.readline()    # skip the initial line with labels
            for line in f:
                a = line.strip().split(',')
                # a goes as follows: model name, chain, index, one letter code, aggrescan score
                x.append(("Chain %s" % a[1], a[2]+a[1]))
                y.append(float(a[-1]))
                name.append(aa_dict_F[a[-2]])
                index.append(a[2])
                chain.append(a[1])
                status.append('Soluble' if float(a[-1]) <= 0 else 'Aggregation prone')
        if not p:
            p = figure(plot_width=1150, plot_height=600, tools=['box_zoom,pan,reset,save'],
                       title="Score breakdown for mutants. "
                             "Click on the legend to hide/show the line. Mouse over a point to see details.",
                       x_range=FactorRange(*x), toolbar_location="below")
            p.xaxis.major_tick_line_color = None  # turn off x-axis major ticks
            p.xaxis.minor_tick_line_color = None  # turn off x-axis minor ticks
            p.yaxis.minor_tick_line_color = None  # turn off x-axis minor ticks
            p.xaxis.major_label_text_font_size = '0pt'  # turn off x-axis tick labels, it keeps the chain names though
            p.xgrid.grid_line_color = None  # turn off the grid
            p.ygrid.grid_line_color = None
        mut_names = [mutant for i in range(len(x))]
        source = ColumnDataSource(data=dict(
            x=x,
            y=y,
            line_y=[0 for i in range(len(x))],
            name=name,
            index=index,
            status=status,
            chain=chain,
            mut_name=mut_names
        ))

        # @ means they will get assigned in order from the data defined above, names means it will only display there
        hover = HoverTool(tooltips=[
            ("Chain", "@chain"),
            ("Residue name", "@name"),
            ("Residue index", "@index"),
            ("Prediction", "@status"),
            ("Mutant", "@mut_name")],
            mode='vline',
            names=[mutant])

        the_line = p.line('x', 'y', source=source, name=mutant,color=colors[counter], line_width=2.5,
                          line_alpha=1.0)  # the main plot
        legend_items.append((mutant, [the_line]))
        if counter not in [0, 1]:
            the_line.visible = False
        p.add_tools(hover)
        counter += 1
    legend = Legend(items=legend_items, click_policy="hide")
    p.add_layout(legend, "left")
    script, div = components(p)
    if save_plot:       # Will not be used for now, could be used by the main program but importing this would be bad
                        # (Due to how the server is started, it would start it on import # TODO provide download button
        save(p, filename="Automated_mutations", title="Mutation analysis")
    else:
        return script, div


def create_plot(csv_address, model):
    x, y, chain, name, index, status = [], [], [], [], [], []
    with open(csv_address, 'r') as f:
        f.readline()    # skip the initial line with labels
        for line in f:
            a = line.strip().split(',')
            # a goes as follows: model name, chain, index, one letter code, aggrescan score
            x.append(("Chain %s" % a[1], a[2]+a[1]))
            y.append(float(a[-1]))
            name.append(aa_dict_F[a[-2]])
            index.append(a[2])
            chain.append(a[1])
            status.append('Soluble' if float(a[-1]) <= 0 else 'Aggregation prone')
    source = ColumnDataSource(data=dict(
        x=x,
        y=y,
        line_y=[0 for i in range(len(x))],
        name=name,
        index=index,
        status=status,
        chain=chain
    ))

    # @ means they will get assigned in order from the data defined above, names means it will only display there
    hover = HoverTool(tooltips=[
        ("Chain", "@chain"),
        ("Residue name", "@name"),
        ("Residue index", "@index"),
        ("Prediction", "@status")],
        mode='vline',
        names=["line"])

    # TODO maybe get the underlying DOM in job_info and adjust the size to that, rather than keeping it static but it wouldnt be perfect either
    # Creeates the plot, adds the tools, and creates the x axis with major tick names, and then chain names below
    p = figure(plot_width=1150, plot_height=600, tools=[hover, 'box_zoom,pan,reset,save'],
               title="Aggrescan3D score based on residue for %s. Mouse over the plot to see residue's details" % model,
               x_range=FactorRange(*x), toolbar_location="below")

    p.xaxis.major_tick_line_color = None  # turn off x-axis major ticks
    p.xaxis.minor_tick_line_color = None  # turn off x-axis minor ticks
    p.yaxis.minor_tick_line_color = None  # turn off x-axis minor ticks
    p.xaxis.major_label_text_font_size = '0pt'  # turn off x-axis tick labels, it keeps the chain names though
    p.xgrid.grid_line_color = None  # turn off the grid
    p.ygrid.grid_line_color = None
    p.line('x', 'y', source=source, name="line")    # the main plot
    glyph = Line(x="x", y="line_y", line_color="#f46d43", line_width=2, line_alpha=0.3)  # small line to help find aggregation prone residues
    p.add_glyph(source, glyph)  # actually draw the line on the figure
    script, div = components(p)

    return script, div



# -*- coding: utf-8 -*-

import os
import errno
import shutil
import logger
from os.path import join, abspath
from subprocess import Popen, PIPE
from glob import glob
from pkg_resources import resource_filename
import json
import tarfile
from postProcessing import prepare_output
from analysis import aggregation_analysis as analyze
from collections import OrderedDict

_name = "CABS"


def run_cabs(config, pdb_input="input.pdb"):
    """Takes an Aggrescan3D Job config dictionary and should never change it"""
    real_work_dir = config["work_dir"]
    os.chdir(real_work_dir)
    real_cabs_dir = os.path.join(real_work_dir, "CABS_sim")
    real_models_dir = os.path.join(real_cabs_dir, "output_pdbs")
    try:
        _makedir(real_cabs_dir)
        _makedir("models")
        _makedir("stats")
        shutil.copyfile(pdb_input, os.path.join(real_cabs_dir,
                                                pdb_input))
    except OSError:
        raise logger.CabsError("Failed to prepare CABS directory at: %s" % real_cabs_dir)

    os.chdir(real_cabs_dir)
    logger.info(module_name=_name,
                msg="Running CABS flex simulation")
    cabs_cmd = _prepare_command(pdb_input=pdb_input, cabs_dir=config["cabs_dir"],
                                cabs_config=config["cabs_config"], n_models=config["n_models"])
    logger.debug(module_name=_name,
                 msg="CABS ran with: %s" % " ".join(cabs_cmd))
    out, err = Popen(cabs_cmd, stdout=PIPE, stderr=PIPE).communicate()
    if err:
        with open(join(real_work_dir, "CABSerror"), 'w') as f:
            f.write(err)
        _cleanup_files(work_dir=real_work_dir, cabs_dir=real_cabs_dir, clean=False)
        raise logger.CabsError("Please see CABSerror file within your work directory for more details",
                               err_file="CABSerror")
    try:
        _check_output(models_dir=real_models_dir, n_models=config["n_models"])
    except logger.CabsError:
        shutil.move(join(real_cabs_dir, "CABS.log"), join(real_work_dir, "CABS.log"))
        _cleanup_files(work_dir=real_work_dir, cabs_dir=real_cabs_dir, clean=False)
        raise logger.CabsError("Please see CABS.log file within your work directory for more details",
                               err_file="CABS.log")

    shutil.copyfile(pdb_input, join("output_pdbs", pdb_input))
    os.chdir("output_pdbs")
    models = glob("model*.pdb")
    top = ""
    max_avg = -100
    averages = {}
    for model in models:
        model_path = abspath(model)
        analyze(config=config, target=model_path, working_dir=real_models_dir, agg_work_dir=real_work_dir)
        stats = prepare_output(work_dir=real_models_dir, final=False,
                               model_name=model.split(".")[0], scores_to_pdb=True)
        current_avg = stats["All"]["avg_value"]
        averages[model] = current_avg
        if current_avg > max_avg:
            max_avg = stats["All"]["avg_value"]
            top = model
        shutil.move("A3D.csv", join(real_work_dir, "stats", model.split(".")[0] + ".csv"))
        shutil.move(model, join(real_work_dir, "models", model))
    analyze(config=config, target=pdb_input, working_dir=real_models_dir, agg_work_dir=real_work_dir)
    stats = prepare_output(work_dir=real_models_dir, final=False,
                           model_name=pdb_input.split(".")[0], scores_to_pdb=True)
    current_avg = stats["All"]["avg_value"]
    averages[pdb_input] = current_avg
    if current_avg > max_avg:
        top = pdb_input
    shutil.move("A3D.csv", join(real_work_dir, "stats", pdb_input.split(".")[0] + ".csv"))
    shutil.move(pdb_input, join(real_work_dir, "models", pdb_input))
    with open('averages', 'w') as avg:
        json.dump(_sort_dict(my_dict=averages), avg)
    os.chdir(real_work_dir)
    _cleanup_files(work_dir=real_work_dir, cabs_dir=real_cabs_dir, input_pdb=pdb_input, top=top, clean=True)
    superimpose(first_model="input.pdb", second_model="folded.pdb")


def superimpose(first_model, second_model):
    try:
        pymol_cmd = ["pymol", "-cq", resource_filename("aggrescan", join("data", "superimpose.pml")), "--", first_model, second_model]
        out, err = Popen(pymol_cmd, stdout=PIPE, stderr=PIPE).communicate()
        if err:
            logger.warning(module_name="Pymol",
                           msg="Pymol reports an error: %s" % err)
        shutil.move("superimposed.png", "CABSflex_supe.png")
    except OSError:
        logger.warning(module_name=_name,
                       msg="Pymol failed to launch (most likely not present on the system)."
                           "Couldn't create a superimposed picture of CABS input and output ")
    except (shutil.Error, IOError):
        logger.critical(module_name="Pymol",
                        msg="Pymol failed to create a superimposed image for input and "
                            "most aggregation prone CABS model")


def _prepare_command(pdb_input="input.pdb", cabs_dir=".", cabs_config='', n_models=12):
    """Prepare CABS settings according to user input"""
    cabs_cmd = []
    if cabs_dir:
        cabs_cmd.extend(["python", cabs_dir, "flex"])
    else:
        cabs_cmd.append("CABSflex")
    if cabs_config:
        cabs_cmd.extend(["-c", cabs_config])
    else:
        cabs_cmd.extend(["--image-file-format", "png", "-v", "4"])
    cabs_cmd.extend(["--input", pdb_input, "--clustering-medoids", str(n_models), "--aa-rebuild", "--log"])
    return cabs_cmd


def _cleanup_files(work_dir="", cabs_dir="", input_pdb="", top="", clean=True):
    """If clean some files will be saved, else only remove all created files"""
    if clean:
        shutil.move(join(cabs_dir, "plots", "RMSF_seq.png"), join(work_dir, "CABSflex_rmsf.png"))
        shutil.move(join(cabs_dir, "plots", "RMSF.csv"), join(work_dir, "CABSflex_rmsf.csv"))
        shutil.copyfile(join(work_dir, "models", top.strip()), join(work_dir, "folded.pdb"))
        shutil.copyfile(join(cabs_dir, "output_pdbs", "averages"), "averages")

    if logger.get_log_level() >= 2 and clean:
        logger.log_file(module_name="CABS",
                        msg="Saving top CABS models as %s" % "models.tar.gz")
        with tarfile.open(join(work_dir, "models.tar.gz"), "w:gz") as tar:
            tar.add(join(work_dir, "models"), arcname=os.path.sep)
        logger.log_file(module_name="CABS",
                        msg="Saving Aggrescan3D statistics for all CABS models as %s" % "stats.tar.gz")
        with tarfile.open(join(work_dir, "stats.tar.gz"), "w:gz") as tar:
            tar.add(join(work_dir, "stats"), arcname=os.path.sep)

    shutil.rmtree(join(work_dir, "stats"), ignore_errors=True)
    shutil.rmtree(join(work_dir, "models"), ignore_errors=True)
    _del_cabs_dir(cabs_dir=cabs_dir)


def _del_cabs_dir(cabs_dir="CABS_sim"):
    shutil.rmtree(cabs_dir, ignore_errors=True)


def _check_output(models_dir, n_models):
    """Check if all the required files were created"""
    _file_list = ["CABS.log"]
    _file_list.extend([join(models_dir, "model_%s.pdb" % str(i)) for i in range(n_models)])
    _file_list.append(join("plots", "RMSF_seq*"))
    _file_list.append(join("plots", "RMSF*"))
    for filename in _file_list:
        if not glob(filename):
            logger.critical(module_name="CABS",
                            msg="File %s which CABS should have generated was not found." % filename)
            raise logger.CabsError


def _sort_dict(my_dict):
    """Return a reverse-sorted by value, OrderedDict of a regular dictionary with number values"""
    new_dict = OrderedDict()
    for key, value in sorted(my_dict.iteritems(), key=lambda (k, v): (v, k), reverse=True):
        new_dict[key] = value
    return new_dict


def _makedir(path):
    """Ignore error if path exists"""
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import pymol
from pymol import cmd, stored, util


pymol_mode = sys.argv[1]
pdb_file = sys.argv[2]
output_file = sys.argv[3]

try:
    pdb_code = pdb_file.split('/')[1].split('.')[0]
except IndexError:
    pdb_code = pdb_file.split('.')[0]
    
if pymol_mode == "s":
    pymol.pymol_argv = ['pymol','-qc'] # PyMol quiet launch without GUI
    out_folder = sys.argv[4]

fh = open(output_file, "r")
lines = fh.readlines()

pymol.finish_launching()
cmd.load(pdb_file)
#cmd.remove("HET")
cmd.set_name(pdb_code, pdb_code+'_A3D')
cmd.hide("all")
cmd.show("surface", "all")
cmd.color("white", "all")

#define the reds
cmd.set_color("red1", [1, 0,   0  ])  # most intense red
cmd.set_color("red2", [1, 0.2, 0.2])
cmd.set_color("red3", [1, 0.4, 0.4])
cmd.set_color("red4", [1, 0.6, 0.6])
cmd.set_color("red5", [1, 0.8, 0.8])  # lightest red

#define the blues
cmd.set_color("blue1", [0,   0,   1])  # most intense blue
cmd.set_color("blue2", [0.2, 0.2, 1])
cmd.set_color("blue3", [0.4, 0.4, 1])
cmd.set_color("blue4", [0.6, 0.6, 1])
cmd.set_color("blue5", [0.8, 0.8, 1])  # lightest blue

high_score = 0
high_res = -1500100900
for line in lines[1:]:
    if not (line == "\n" or line.startswith('#')):
        line_params = line.strip().replace('//', '').split(" ")
        score = float(line_params[3])

        chain = line_params[0]
        if chain == '-':
            my_res = "resi " + line_params[1]
        else:
            my_res = "chain " + line_params[0] + " and resi " + line_params[1]

        if score > 0:  # residues with a value above 0 will be colored red
            if score < 0.5:
                cmd.color("red4", my_res)    
            elif score < 1:
                cmd.color("red3", my_res)
            elif score < 1.5:
                cmd.color("red2", my_res)
            else:
                cmd.color("red1", my_res)
            if score > high_score:
                high_score = score
                high_res = my_res
        elif score < -0.5:  # residues with a value below -0.5 will be colored blue
            if score > -1:
                cmd.color("blue5", my_res)
            elif score > -1.5:
                cmd.color("blue4", my_res)
            elif score > -2:
                cmd.color("blue3", my_res)
            elif score > -2.5:
                cmd.color("blue2", my_res)
            else:
                cmd.color("blue1", my_res)

# create an object for the high score hot spot of the protein
if high_res != -1500100900:
    cmd.create("hot_spot", "("+high_res+",byres "+high_res+" around 5 and not HET \
    and not color white and not color blue1 and not color blue2 and not color blue3 \
    and not color blue4 and not color blue5)")

stored.neighbours=[]
cmd.iterate("(hot_spot & n. ca)", "stored.neighbours.append((resi,resn))")
cmd.reset()


# Save PyMol session
if pymol_mode == "s":
    cmd.viewport("640", "480")
    cmd.origin()
    cmd.set("field_of_view",30)
    cmd.set("surface_quality", "1")
    cmd.set("ray_trace_mode", "1")
    cmd.bg_color("white")
    cmd.set("antialias", "2")
    for xx in range(1, 19):
	cmd.rotate("y", 20)
        cmd.ray("640", "480")
	cmd.png("mov%05d.png" % (xx))


    cmd.quit()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys,shutil
import logger
from newRunJob import Job
import optparser as opt
import traceback as _tr

_name = "Main"


def _cleanup(aggrescan_job_instance):
    """
    Done to prevent a situation where Job raises an error on initialization and there is nothing to delete
    Normally deletes or keeps the simulations temporary directory depending on verbosity settings
    (Can't check for Exceptions in finally block, hence the function)
    """
    try:
        if logger.get_log_level() < 4:
            shutil.rmtree(aggrescan_job_instance.get_tempdir(),ignore_errors=True)
            logger.log_file(module_name=_name, msg="Removing temporary files")
        else:
            logger.log_file(module_name=_name,
                            msg="Verbosity higher than 3 - temporary files kept in %s"
                                % aggrescan_job_instance.get_tempdir())
    except AttributeError:
        pass


def run_program():
    try:
        a = "dummy"  # to avoid cleanup on empty instances
        options = opt.parse(options=sys.argv[1:])
        logger.setup(log_level=options['verbose'], remote=options['remote'], work_dir=options["work_dir"])
        a = Job(config=options)
        a.run_job()
        _cleanup(a)
        logger.info(module_name="Main", msg="Simulation completed successfully.")

    except KeyboardInterrupt:
        _cleanup(a)
        logger.info(module_name=_name, msg="Interrupted by user")

    except logger.AggrescanError as custom_error:
        _cleanup(a)
        custom_error.generate_error_file()
        logger.exit_program(module_name=custom_error.module_name,
                            msg=custom_error.logger_msg,
                            traceback=None,
                            exc=custom_error)

    except Exception as e:
        _cleanup(a)
        logger.record_exception(trace_stack=_tr.format_exc())
        logger.critical(module_name=_name,
                        msg="Unhandled Exception caught: %s." % e.message)
        if logger.get_log_level() > 2:
            logger.info(module_name=_name,
                        msg="Verbosity higher than 2 - raising the Exception to provide traceback.")
            raise
        else:
            logger.exit_program()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import subprocess
import sys
flask_loc = os.path.dirname(os.path.abspath(__file__))
try:
    from flask import Flask
    app = Flask(__name__)
    config = dict(PAGINATION=45,

                  STATIC=os.path.join(flask_loc, 'static'),
                  PICTURES=os.path.join(flask_loc, 'pictures'),
                  USERJOB_DIRECTORY=os.path.join(flask_loc, 'computations'),
                  DEBUG=True,
                  SECRET_KEY="This_apps_super_secret_key")

    app.config.update(config)

    import viewsutils
    import views
    import db_manager
    import job_handling


    # Create an empty database for the user only if one doesn't exist.
    if not os.path.isfile(os.path.join(flask_loc, 'a3d_database.db')):
        db_manager.create_new_database(os.path.join(flask_loc, 'a3d_database.db'))
        print "Database for your projects has been created."

    app.config.update(dict(DATABASE=os.path.join(flask_loc, 'a3d_database.db')))


except ImportError:
    with open(os.path.join(flask_loc, "requirements.txt"), 'r') as f:
        print "This program requires a few python packages to run: "
        for line in f:
            print line.strip()
        print "Do you wish to use pip to install/upgrade those?"
    test = raw_input("Type 'y' or 'Y' if yes, else press enter to quit\n")
    if test == 'y' or test == 'Y':
        subprocess.Popen(["pip", "install", "-r", os.path.join(flask_loc, "requirements.txt")]).communicate()
    else:
        print "The program has failed to start. Please contact us if you need further assistance."
        sys.exit(0)
    print ""
    print "The program attempted to install necessary packages. Re-run this script to run the app."
    print "Note, the script will fail if you don't have pip installed. If that is the case please" \
          " visit https://pip.pypa.io/en/stable/installing/ for instructions on how to install it."
    print ""




#!/usr/bin/env python
# -*- coding: utf-8 -*-


import json
import shutil
import sqlite3
from flask import render_template, url_for, request, flash, redirect, jsonify, g, abort, Response

from flask_wtf import Form
from flask_wtf.file import FileField, FileAllowed
from wtforms import StringField, BooleanField, HiddenField, RadioField
from wtforms.validators import Length, Email, optional, ValidationError

from a3d_gui import app
from utils import query_db, unique_id, gunzip, connect_db, parse_out, aminoacids, get_undefined_job_ids
from job_handling import run_job, check_jobs
from create_config import generate_config, prepare_data
from validators import *

##############################################################################


@app.before_request
def before_request():
    g.sqlite_db = connect_db()


def flash_errors(form):
    for field, errors in form.errors.items():
        for error in errors:
            flash(u"Error in the %s field - %s\n    " %
                  (getattr(form, field).label.text, error), 'error')


class MyForm(Form):
    name = StringField('Project name',
                       validators=[Length(min=4, max=50), optional()],
                       description="Project name",
                       default="")
    input_pdb_code = StringField('PDB code',
                                 validators=[pdb_input_validator,
                                             pdb_code_validator],
                                 description="1YWO",
                                 default="")
    chain = StringField('Chain',
                        validators=[Length(min=1, max=5),
                                    optional(), single_chain_validator],
                        description="A",
                        default="")
    input_file = FileField('Local PDB file',
                           validators=[pdb_file_validator,
                                       non_empty_file])
    jid = HiddenField(default=unique_id())
    foldx = RadioField("Stability calculations",
                       validators=[foldx_validator],
                       choices=[('1', 'Yes'), ('0', 'No')], default='1')
    mutation = RadioField('Mutate residues',
                          choices=[('1', 'Yes'), ('0', 'No')],
                          validators=[foldx_and_m_validator],
                          default='0')
    auto_mutation = RadioField('Enhance protein solubility',
                               choices=[('1', 'Yes'), ('0', 'No')],
                               validators=[auto_mut_validator],
                               default='0')
    dynamic = RadioField("Dynamic mode",
                         choices=[('1', 'Yes'), ('0', 'No')], default='0')
    distance = RadioField('Distance of aggregation analysis: ',
                          choices=[('5', '5Å'.decode('utf-8')),
                                   ('10', '10Å'.decode('utf-8'))],
                          default='10')


def add_init_data_to_db(form):
    jid = unique_id()
    name = form.name.data
    if len(name) < 2:
        name = jid
    chains = form.chain.data.strip()
    dest_directory = os.path.abspath(os.path.join(app.config['USERJOB_DIRECTORY'], jid))
    dest_file = os.path.join(dest_directory, "input.pdb.gz")
    os.makedirs(dest_directory)
    if form.input_pdb_code.data:
        pdb_code = form.input_pdb_code.data.strip()
        stream = urllib2.urlopen('http://www.rcsb.org/pdb/files/' + pdb_code + '.pdb.gz')
        b2 = stream.read()
        string_io_data = StringIO(b2)
        with gzip.GzipFile(fileobj=string_io_data, mode="rb") as f:
            p = PdbParser(f, chains)
            input_seq = p.getSequence()
            p.savePdbFile(dest_file)
            gunzip(dest_file)
        stream.close()
    else:
        p = PdbParser(form.input_file.data.stream, chains)
        input_seq = p.getSequence()
        p.savePdbFile(dest_file)
        gunzip(dest_file)

    chain_numbering = p.getChainIdxResname()
    distance = form.distance.data
    dynamic = form.dynamic.data
    mutate = form.mutation.data
    foldx = form.foldx.data
    all_chains = chains or "".join(p.getChains())
    if mutate == "1":
        query_db("INSERT INTO user_queue (jid, project_name, working_dir, status)\
                VALUES(?,?,?,?)",
                 [jid, name, dest_directory, 'mut_wait'], insert=True)
        query_db("INSERT INTO project_details (jid, chain_numbering, chain_sequence, \
                 dynamic, mutate, distance, chains, foldx) \
                 VALUES(?,?,?,?,?,?,?,?)",
                 [jid, chain_numbering, input_seq, dynamic, mutate, distance, all_chains, foldx], insert=True)
    else:
        query_db("INSERT INTO user_queue (jid, project_name, working_dir, status)\
                VALUES(?,?,?,?)",
                 [jid, name, dest_directory, 'running'], insert=True)
        query_db("INSERT INTO project_details (jid, chain_numbering, chain_sequence, \
                 dynamic, mutate, distance, chains, foldx) \
                 VALUES(?,?,?,?,?,?,?,?)",
                 [jid, chain_numbering, input_seq, dynamic, mutate, distance, all_chains, foldx], insert=True)

    return jid, dest_directory


@app.route('/job/<jid>/')
def job_status(jid):
    try:
        all_info = prepare_data(jid)
    except IOError as e:
        flash("Your job tried to access files that don't exists. Perhaps a job with status done had its files removed? "
              "Or you tried to set done on a job that misses aggrescan files.  "
              "Changed status to error.")
        flash("Your error: %s" % str(e))
        query_db("UPDATE user_queue SET status='error'\
                  WHERE jid=?", [jid], insert=True)
        return job_status(jid)
    aa_t = aminoacids()
    return render_template('job_info.html',
                           jid=jid, aa_dict=aa_t, info=all_info)


@app.route('/mutate/<jid>', methods=['GET', 'POST'])
def mutate(jid):
    q = query_db("SELECT chain_numbering FROM project_details \
                  WHERE jid=?", [jid], one=True)
    chains = json.loads(q['chain_numbering'])
    aa_t = aminoacids()

    return render_template('mutate.html', sequence=chains,
                           jid=jid, aa_dict=aa_t)


@app.route('/auto_mutate/<jid>', methods=['GET', 'POST'])
def auto_mutate(jid):
    q = query_db("SELECT chain_numbering FROM project_details \
                  WHERE jid=?", [jid], one=True)
    chains = json.loads(q['chain_numbering'])
    aa_t = aminoacids()
    return render_template('auto_mutate.html', sequence=chains,
                           jid=jid, aa_dict=aa_t)


@app.route('/_upmut2', methods=['GET', 'POST'])
def up_remutation():
    if request.method == 'POST':
        jid = unique_id()
        old_jid = request.form.get('oldjid', '')
        mut = ", ".join(request.form.getlist('toreplace[]'))
        old_queue = query_db("SELECT project_name, working_dir FROM \
                user_queue WHERE jid=?", [old_jid], one=True)
        new_name = old_queue['project_name']+" [mutate: "+str(mut)+"]"

        d = query_db("SELECT foldx, dynamic, distance, chains FROM project_details WHERE jid=?",
                     [old_jid], one=True)
        dest_directory = os.path.join(app.config['USERJOB_DIRECTORY'], jid)
        old_directory = old_queue['working_dir']
        os.mkdir(dest_directory)
        shutil.copy(os.path.join(old_directory, "output.pdb"),
                    os.path.join(dest_directory, "input.pdb"))
        with open(os.path.join(dest_directory, "input.pdb"), "rb") as rea:
            p = PdbParser(rea)
            sequence = p.getSequence()
            chain_numbering = p.getChainIdxResname()

        query_db("INSERT INTO user_queue(jid, project_name, working_dir, status) VALUES(?,?,?,?)",
                 [jid, new_name, dest_directory, 'running'], insert=True)
        query_db("INSERT INTO project_details(jid, dynamic, mutate, mutation, chains, foldx, distance, "
                 "chain_sequence, chain_numbering) VALUES(?,?,?,?,?,?,?,?,?)",
                 [jid, d['dynamic'], 1, mut, d['chains'], 1, d['distance'],
                  sequence, chain_numbering], insert=True)
        generate_config(jid)
        pid = run_job(dest_directory)
        query_db("UPDATE user_queue SET pid=? WHERE jid=?", [pid, jid], insert=True)
        return redirect(url_for('job_status', jid=jid))


@app.route('/_upmut', methods=['GET', 'POST'])
def up_mutation():
    if request.method == 'POST':
        jid = request.form.get('jid', '')
        mut = ", ".join(request.form.getlist('toreplace[]'))
        details = query_db("SELECT project_name, working_dir FROM user_queue WHERE \
                jid=?", [jid], one=True)
        proname = details['project_name'] + " [mutate: "+mut+"]"
        query_db("UPDATE user_queue SET status='running', \
                project_name=? WHERE jid=?", [proname, jid], insert=True)
        query_db("UPDATE project_details SET mutate=?, mutation=? WHERE jid=?", [1, mut, jid], insert=True)
        generate_config(jid)
        pid = run_job(details['working_dir'])
        query_db("UPDATE user_queue SET pid=? WHERE jid=?", [pid, jid], insert=True)
        return redirect(url_for('job_status', jid=jid))


@app.route('/_upautomut', methods=['GET', 'POST'])
def up_automutation():
    if request.method == 'POST':
        jid = request.form.get('jid', '')
        n_mutated = request.form.get('n_mutated', '')
        n_cores = request.form.get('n_cores', '')

        automut = "%s %s " % (n_mutated, n_cores) + " ".join(request.form.getlist('toreplace[]'))
        details = query_db("SELECT project_name, working_dir FROM user_queue WHERE \
                jid=?", [jid], one=True)
        proname = details['project_name'] + " (automated mutations)"
        query_db("UPDATE user_queue SET status='running', \
                project_name=? WHERE jid=?", [proname, jid], insert=True)
        try:
            query_db("UPDATE project_details SET auto_mutation=? WHERE jid=?", [automut, jid], insert=True)
        except sqlite3.OperationalError: # Backwards compatibility when the column didn't exist
            query_db("ALTER TABLE project_details ADD COLUMN auto_mutation TEXT", insert=True)
            query_db("UPDATE project_details SET auto_mutation=? WHERE jid=?", [automut, jid], insert=True)
        generate_config(jid)
        pid = run_job(details['working_dir'])
        query_db("UPDATE user_queue SET pid=? WHERE jid=?", [pid, jid], insert=True)
        return redirect(url_for('job_status', jid=jid))


@app.route('/_check_status/<jid>', methods=['GET'])
def get_job_status(jid):
    response = {}
    details = query_db("SELECT working_dir,status,jid FROM user_queue WHERE \
            jid=?", [jid], one=True)
    try:
        with open(os.path.join(details['working_dir'], "Aggrescan.log"), 'r') as f:
            response['log'] = f.read()
        if (details['status'] != 'done' and details['status'] != 'error') and check_jobs(details['jid']):
            response['reload'] = True
    except IOError:
        response['log'] = "Couldn't open the log for simulation at %s" % os.path.join(details['working_dir'], "Aggrescan.log")
    return jsonify(**response)


@app.route('/index', methods=['GET', 'POST'])
@app.route('/', methods=['GET', 'POST'])
def index_page():
    form = MyForm()
    foldx_path = query_db("SELECT foldx_path FROM system_data WHERE id=1", one=True)[0]
    if request.method == 'POST':
        if form.validate_on_submit():
            jid, work_dir = add_init_data_to_db(form)
            if form.mutation.data == '1':
                return redirect(url_for('mutate', jid=jid))
            elif form.auto_mutation.data == '1':
                return redirect(url_for("auto_mutate", jid=jid))
            else:
                generate_config(jid)
                pid = run_job(work_dir)
                query_db("UPDATE user_queue SET pid=? WHERE jid=?", [pid, jid], insert=True)
                return redirect(url_for('job_status', jid=jid))
        flash_errors(form)
    return render_template('index.html', form=form, foldx_path=foldx_path)


########################################################################################################################
@app.route('/queue', methods=['POST', 'GET'], defaults={'page': 1})
@app.route('/queue/page/<int:page>/', methods=['POST', 'GET'])
def queue_page(page=1):
    before = (page - 1) * app.config['PAGINATION']
    if request.method == 'GET':
        search = request.args.get('q', '')
        if search != '':
            flash("Searching results for %s ..." % (search), 'warning')
            q = query_db("SELECT project_name, jid, status, \
                    started datet FROM user_queue \
                    WHERE status!='mut_wait' AND (project_name LIKE ? OR \
                    jid=?) ORDER BY started DESC LIMIT ?,?",
                         ["%"+search+"%", search, before,
                          app.config['PAGINATION']])
            q_all = query_db("SELECT status FROM user_queue WHERE \
                    status!='mut_wait' AND (project_name LIKE ? OR jid=?) \
                    ORDER BY started DESC", ["%"+search+"%", search])
            out = parse_out(q)
            if len(out) == 0:
                flash("Nothing found", "error")
            elif len(out) == 1:
                flash("Project found!", "info")
                jid = out[0]['jid']
                return redirect(url_for('job_status', jid=jid))

            return render_template('queue.html', queue=out, page=page,
                                   total_rows=len(q_all))
    # Get the jobs from database
    qall = query_db("SELECT status FROM user_queue WHERE \
            status!='mut_wait' ORDER BY started DESC", [])
    q = query_db("SELECT jid, project_name, status, started datet \
                  FROM user_queue WHERE status != 'mut_wait' \
                  ORDER BY started DESC LIMIT ?,?", [before, app.config['PAGINATION']])
    # Check if status has changed
    check_jobs(*get_undefined_job_ids(q))
    # Get updated que
    q = query_db("SELECT jid, project_name, status, started datet \
                      FROM user_queue WHERE status != 'mut_wait' \
                      ORDER BY started DESC LIMIT ?,?", [before, app.config['PAGINATION']])
    out = parse_out(q)
    return render_template('queue.html', queue=out, total_rows=len(qall),
                           page=page)








#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys

# Make sure the path to actual files is on the path (the script pretends to be in the parent directory this way)
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(os.path.abspath(__file__)), os.pardir)))


def run_server():
    try:
        from a3d_gui import app
        app.run(host='0.0.0.0', debug=True)
    except ImportError:     # The exception is caught in __init__ but this script raises its own when the import fails
        pass                # Or smth I might not know what exactly happens



from libc.stdio cimport FILE

cdef extern from "freesasa.h":
    ctypedef enum freesasa_algorithm:
        FREESASA_LEE_RICHARDS, FREESASA_SHRAKE_RUPLEY

    ctypedef enum freesasa_verbosity:
        FREESASA_V_NORMAL, FREESASA_V_NOWARNINGS, FREESASA_V_SILENT, FREESASA_V_DEBUG

    ctypedef enum freesasa_atom_class:
        FREESASA_ATOM_APOLAR, FREESASA_ATOM_POLAR, FREESASA_ATOM_UNKNOWN

    cdef int FREESASA_SUCCESS
    cdef int FREESASA_FAIL
    cdef int FREESASA_WARN

    cdef int FREESASA_INCLUDE_HETATM
    cdef int FREESASA_INCLUDE_HYDROGEN
    cdef int FREESASA_SEPARATE_CHAINS
    cdef int FREESASA_SEPARATE_MODELS
    cdef int FREESASA_JOIN_MODELS
    cdef int FREESASA_HALT_AT_UNKNOWN
    cdef int FREESASA_SKIP_UNKNOWN

    cdef int FREESASA_MAX_SELECTION_NAME

    ctypedef struct freesasa_parameters:
        freesasa_algorithm alg
        double probe_radius
        int shrake_rupley_n_points
        int lee_richards_n_slices
        int n_threads

    ctypedef struct freesasa_result:
        double total
        double *sasa
        int n_atoms

    ctypedef struct freesasa_nodearea:
        const char *name
        double total
        double main_chain
        double side_chain
        double polar
        double apolar
        double unknown

    ctypedef struct freesasa_classifier:
        pass

    ctypedef struct freesasa_structure:
        pass

    ctypedef struct freesasa_selection:
        pass

    cdef extern const freesasa_parameters freesasa_default_parameters
    cdef extern const freesasa_classifier freesasa_default_classifier
    cdef extern const freesasa_classifier freesasa_residue_classifier

    freesasa_result* freesasa_calc_structure(const freesasa_structure *structure,
                                             const freesasa_parameters *parameters)

    freesasa_result* freesasa_calc_coord(const double *xyz,
                                         const double *radii,
                                         int n,
                                         const freesasa_parameters *parameters)

    void freesasa_result_free(freesasa_result *result)

    freesasa_classifier* freesasa_classifier_from_file(FILE *file)

    void freesasa_classifier_free(freesasa_classifier *classifier)

    int freesasa_structure_chain_residues(const freesasa_structure *structure,
                                          char chain,
                                          int *first,
                                          int *last)

    double freesasa_classifier_radius(const freesasa_classifier *classifier,
                                      const char *res_name,
                                      const char *atom_name)

    freesasa_atom_class freesasa_classifier_class(const freesasa_classifier *classifier,
                                                  const char *res_name,
                                                  const char *atom_name)

    const char* freesasa_classifier_class2str(freesasa_atom_class the_class)

    freesasa_selection * freesasa_selection_new(const char *command,
                                                const freesasa_structure *structure,
                                                const freesasa_result *result)

    void freesasa_selection_free(freesasa_selection *selection)

    const char * freesasa_selection_name(const freesasa_selection* selection)

    const char * freesasa_selection_command(const freesasa_selection* selection)

    double freesasa_selection_area(const freesasa_selection* selection)

    int freesasa_selection_n_atoms(const freesasa_selection* selection)

    int freesasa_write_pdb(FILE *output,
                           freesasa_result *result,
                           const freesasa_structure *structure)

    int freesasa_per_residue_type(FILE *output,
                                  freesasa_result *result,
                                  const freesasa_structure *structure)

    int freesasa_per_residue(FILE *output,
                             freesasa_result *result,
                             const freesasa_structure *structure)

    int freesasa_set_verbosity(freesasa_verbosity v)

    freesasa_verbosity freesasa_get_verbosity()

    freesasa_structure* freesasa_structure_from_pdb(FILE *pdb,
                                                    const freesasa_classifier* classifier,
                                                    int options)

    freesasa_structure** freesasa_structure_array(FILE *pdb,
                                                  int *n,
                                                  const freesasa_classifier* classifier,
                                                  int options)

    freesasa_structure* freesasa_structure_new()

    int freesasa_structure_n(freesasa_structure *structure)

    void freesasa_structure_free(freesasa_structure* structure)

    const double* freesasa_structure_radius(const freesasa_structure *structure)

    void freesasa_structure_set_radius(freesasa_structure *structure,
                                       const double *radii)

    
    int freesasa_structure_add_atom(freesasa_structure *structure,
                                    const char* atom_name,
                                    const char* residue_name,
                                    const char* residue_number,
                                    char chain_label,
                                    double x, double y, double z)

    int freesasa_structure_add_atom_wopt(freesasa_structure *structure,
                                         const char* atom_name,
                                         const char* residue_name,
                                         const char* residue_number,
                                         char chain_label,
                                         double x, double y, double z,
                                         const freesasa_classifier *classifier,
                                         int options)

    const char* freesasa_structure_atom_name(const freesasa_structure *structure,
                                             int i)

    const char* freesasa_structure_atom_res_name(const freesasa_structure *structure,
                                                 int i)

    const char* freesasa_structure_atom_res_number(const freesasa_structure *structure,
                                                   int i)

    double freesasa_structure_atom_radius(const freesasa_structure *structure,
                                          int i)

    void freesasa_structure_atom_set_radius(const freesasa_structure *structure,
                                            int i,
                                            double radius)

    char freesasa_structure_atom_chain(const freesasa_structure *structure, int i)

    const double* freesasa_structure_coord_array(const freesasa_structure *structure)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from a3d_gui import app
from utils import query_db
from server_plot import create_plot
import os
import tarfile
import psutil
import shutil
import sqlite3
import signal
from os.path import join, isfile, exists
from subprocess import Popen, PIPE
from glob import glob
from flask import Response, flash, url_for


"""
Module intended for job (process, db status) managing for the Aggrescan3D standalone app. 
CAUTION! Some of the functions here perform file deletion as well as process killing                        
Should you chose to modify basic settings or job handling and db managing functions do so with care 
"""

_static_file_list = set(["A3D.csv", "output.pdb"])
_dynamic_file_list = set(["CABSflex_rmsf.png", "CABSflex_rmsf.csv", "averages"])
_auto_mut_file_list = set(["A3D.csv", "output.pdb", "Mutations_summary.csv"])
_other_files = set(["Aggrescan.log", "Aggrescan.error", "input.pdb", "input.pdb.gz", "clip.webm", "models.tar.gz",
                    "stats.tar.gz", "config.ini", "output.pdb", "MutantEnergyDiff"])


def run_job(working_dir):
    curr_dir = os.path.abspath(os.curdir)      # save the working dir of server app, otherwise it loses track of the db
    try:
        os.chdir(working_dir)
        process = Popen(["aggrescan", "-c", "config.ini"],
                        stdout=PIPE,
                        stderr=PIPE,
                        preexec_fn=os.setpgrp)    # hopefully the prevents the job from dying with the server
        return process.pid
    except AttributeError:  # Windows band-aid
        os.chdir(working_dir)
        process = Popen(["aggrescan", "-c", "config.ini"],
                        stdout=PIPE,
                        stderr=PIPE)
        return process.pid
    finally:
        os.chdir(curr_dir)  # It is crucial that python stays within its starting directory


@app.route('/_rerun_job/<jid>', methods=['GET'])
def rerun_job(jid):
    working_dir = query_db("SELECT working_dir FROM user_queue WHERE jid=?", [jid])[0][0]
    job_pid = query_db("SELECT pid FROM user_queue WHERE jid=?", [jid])[0][0]
    if isfile(join(working_dir, "config.ini")):
        if is_running(job_pid):
            return Response("The job appears to still be running. Stop it first. Re-run aborted.",
                            status=400, content_type="plain/text")
        try:
            os.remove(join(working_dir, "Aggrescan.log"))
        except OSError:
            pass
        new_pid = run_job(working_dir)
        query_db("UPDATE user_queue SET status='running', pid=? WHERE jid=?", [new_pid, jid], insert=True)
        return url_for('job_status', jid=jid)
    else:
        return Response("File config.ini not present in working directory for the job. Re-run aborted.",
                        status=400, content_type="plain/text")


@app.route('/_kill_job/<jid>', methods=['GET'])
def kill_job(jid):
    """Kill the aggrescan process and all its children, set status to error"""
    proc_pid = query_db("SELECT pid FROM user_queue WHERE jid=?", [jid])[0][0]
    try:
        process = psutil.Process(proc_pid)
        if process.name() == "aggrescan" and process.is_running():
            kill_proc_tree(proc_pid)
            query_db("UPDATE user_queue SET status='error' WHERE jid=?", [jid], insert=True)
            return Response("Job killed", status="200", content_type="plain/text")
        else:
            return Response("Coulnd't stop the job. The pid assigned to it no "
                            "longer points to an active aggrescan job.",
                            status="400", content_type="plain/text")
    except psutil.NoSuchProcess:
        return Response("Coulnd't stop the job. There is no such process.",
                        status="400", content_type="plain/text")
    except psutil.AccessDenied as e:
        return Response("You dont have permissions to stop the job. %s" % str(e),
                        status="400", content_type="plain/text")
    except psutil.TimeoutExpired as e:
        return Response("Coulnd't stop the job. %s" % str(e),
                        status="400", content_type="plain/text")
    except ValueError as e:
        return Response("Can't kill this job. "
                        "Most likely this job was started manually and can't be stopped by this interface",
                        status="400", content_type="plain/text")


@app.route('/_kill_delete_job/<jid>', methods=['GET'])
def kill_delete_job(jid):
    """Terminate the aggrescan process and all its children, and delete all the aggrescan files"""
    try:
        kill_response = kill_job(jid)
        delete_response = delete_files(jid)
        if delete_response and kill_response.status_code == 200:
            _silent_delete_job(jid)
            flash("Files and directory deleted. The job %s has been stopped." % jid)
            return url_for('index_page')
        elif not delete_response and kill_response.status_code == 200:
            _silent_delete_job(jid)
            flash("Files deleted but directory still contains files. The job %s has been stopped." % jid)
            return url_for('index_page')
        elif delete_response and kill_response.status_code != 200:
            _silent_delete_job(jid)
            flash("Something went wrong. The files and directory were deleted but the job couldn't be stopped."
                  "Reason: %s" % kill_response.response)
            return url_for('index_page')
        else:
            return Response("Something went wrong. The directory was not deleted"
                            "and the job couldn't be stopped."
                            "Reason: %s" % kill_response.response,
                            status="400", content_type="plain/text")
    except Exception as e:  # Not nice but there is no outer instance to take that to and I really don't want this crashing the app
        return Response("Something went terribly wrong and your request failed in an unpredictable way ",
                        status="400", content_type="plain/text")


def is_running(proc_pid):
    """Find the aggrescan process based on database-stored pid"""
    try:
        process = psutil.Process(proc_pid)
        if process.name() == "aggrescan" and process.is_running() and process.status() != "zombie":
            return True
        return False
    except psutil.Error:
        return False
    except ValueError:  # for negative pids on added jobs
        return False


def check_jobs(*jobs):
    """
    Attempt to check the job's status. Whether the job is done, missing files, an error occurred, etc
    :param jobs: a list of job_ids
    :return: Bool indicating if the job is done but only if one job id was provided (if more - always None)
    """
    for jid in jobs:
        done = False    # helper to reduce the number of ifs
        working_dir = query_db("SELECT working_dir FROM user_queue WHERE jid=?", [jid])[0][0]
        job_pid = query_db("SELECT pid FROM user_queue WHERE jid=?", [jid])[0][0]
        if isfile(join(working_dir, "Aggrescan.error")):
            query_db("UPDATE user_queue SET status='error' WHERE jid=?", [jid], insert=True)
        elif is_running(job_pid):
            query_db("UPDATE user_queue SET status='running' WHERE jid=?", [jid], insert=True)
        elif not exists(working_dir):
            query_db("UPDATE user_queue SET status='missing_files' WHERE jid=?", [jid], insert=True)
        else:
            details = query_db("SELECT dynamic, mutate, auto_mutation FROM project_details WHERE jid=?", [jid], one=True)
            files_in_dir = set([os.path.basename(i) for i in glob(join(working_dir, "*"))])
            if details['dynamic']:  # attempt to deduct if the job is running or the files are simply not present in the working dir
                missing_files = len(_static_file_list - files_in_dir) + len(_dynamic_file_list - files_in_dir)
                if missing_files == 0:
                    query_db("UPDATE user_queue SET status='done' WHERE jid=?", [jid], insert=True)
                    prepare_files(working_dir)
                    done = True
                else:
                    query_db("UPDATE user_queue SET status='missing_files' WHERE jid=?", [jid], insert=True)

            elif details['auto_mutation']:
                if _auto_mut_file_list - files_in_dir:  # "since the most important files are present we assume the job is done"
                    query_db("UPDATE user_queue SET status='missing_files' WHERE jid=?", [jid], insert=True)
                else:
                    query_db("UPDATE user_queue SET status='done' WHERE jid=?", [jid], insert=True)
                    done = True
            else:
                if _static_file_list - files_in_dir:  # "since two most important files are present we assume the job is done"
                    query_db("UPDATE user_queue SET status='missing_files' WHERE jid=?", [jid], insert=True)
                else:
                    query_db("UPDATE user_queue SET status='done' WHERE jid=?", [jid], insert=True)
                    done = True
            if details['mutate'] and done:
                try:
                    with open(join(working_dir, "MutantEnergyDiff"), 'r') as f:
                        mut_energy_dif = f.read().split()[0].strip()
                    query_db("UPDATE project_details SET mutt_energy_diff=? WHERE jid=?",
                             [mut_energy_dif, jid], insert=True)
                except IOError:
                    query_db("UPDATE user_queue SET status='missing_files' WHERE jid=?", [jid], insert=True)
    if len(jobs) == 1:
        return done


def prepare_files(working_dir):
    """Move some files around to make it easier for the pseudo server"""
    curr_dir = os.path.abspath(os.curdir)      # save the working dir of server app, otherwise it loses track of the db
    try:    # make absolutely sure it gets back to the curr_dir
        os.chdir(working_dir)
        with tarfile.open("models.tar.gz", 'r:gz') as tar:
            tar.extractall()
        os.remove("models.tar.gz")  # remove the archive since the files are already outside anyway
    except IOError:
        pass  # models are not present in the folder, this can happen shouldn't be an issue
    try:
        os.chdir(working_dir)
        with tarfile.open("stats.tar.gz", 'r:gz') as tar:
            tar.extractall()
        os.remove("stats.tar.gz")  # remove the archive since the files are already outside anyway
    except IOError:
        pass  # stats are not present in the folder, this can happen shouldn't be an issue
    finally:
        os.chdir(curr_dir)


def delete_files(job_id):
    """
    Delete all aggrescan files in the working directory. If the directory is then empty it is also removed
    Warning! This will delete any files and subfolders in "tmp" folder in your working directory.
    Returns True if the entire dir is deleted, False otherwise
    """
    n_models = 12  # TODO customizable number of models for the gui
    working_dir = query_db("SELECT working_dir FROM user_queue WHERE jid=?", [job_id])[0][0]
    chains = query_db("SELECT chains FROM project_details WHERE jid=?", [job_id])[0][0]
    pictures = query_db("SELECT * FROM pictures WHERE jid=?", [job_id])
    if pictures is not None:
        for pic in pictures:
            try:
                os.remove(join(app.config['PICTURES'], pic['filename']))
            except OSError:
                pass  # the picture was likely deleted from the folder manually
    files = ["%s.png" % i for i in chains]
    files.extend(["%s.svg" % i for i in chains])
    files.extend(["model_%d.pdb" % i for i in range(n_models)])
    files.extend(["model_%d.pdb" % i for i in range(n_models)])
    files_in_dir = [i for i in glob(working_dir+"/*") if isfile(i)]
    shutil.rmtree(join(working_dir, "tmp"), ignore_errors=True)
    for to_del in files_in_dir:
        try:
            os.remove(to_del)
        except OSError:
            flash("Couldn't delete a file, which was supposed to be here: %s" % to_del)
    try:
        if os.listdir(working_dir):
            return False
        else:
            try:
                os.rmdir(working_dir)
                return True
            except OSError:
                return False  # Something went in the way, shouldn't happen but if it does lets not make a mess about it
    except OSError:
        flash("Attempted to delete %s but it already doesn't exist." % working_dir)
        return True


def kill_proc_tree(pid, sig=signal.SIGTERM, include_parent=True,
                   timeout=3):
    """
    Kill a process tree. SIGTERM is sent first, then SIGKILL.
    If a process survives that a psutil.TimeoutExpired is raised
    If everything goes well None is the return value
    """
    if pid == os.getpid():
        raise RuntimeError("I refuse to kill myself")
    parent = psutil.Process(pid)
    children = parent.children(recursive=True)
    if include_parent:
        children.append(parent)
    for p in children:
        try:
            p.send_signal(sig)
        except psutil.NoSuchProcess:    # Parent likely dies on child death so just ignore that
            pass
    gone, alive = psutil.wait_procs(children, timeout=timeout)
    if alive:
        for p in alive:
            p.send_signal(signal.SIGKILL)
    gone, alive = psutil.wait_procs(children, timeout=timeout)
    if alive:
        raise psutil.TimeoutExpired("Failed to stop or kill a process")


def _silent_delete_job(jid):
    """Delete a job from current database, returns only success or failure"""
    try:
        query_db("DELETE FROM user_queue WHERE jid=?", [jid], insert=True)
        query_db("DELETE FROM project_details WHERE jid=?", [jid], insert=True)
    except sqlite3.Error:
        flash("For some reason couldn't delete %s from database. "
              "Perhaps it already doesn't exist?" % jid, 'alert')




#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import shutil
import json
from pkg_resources import resource_filename
from subprocess import Popen, PIPE
from os.path import exists, isdir, join, isfile
import logger
import pdb
from postProcessing import prepare_output
from analysis import aggregation_analysis as analyze
from dynamic_module import run_cabs
from foldx_module import FoldxWrap as fold
from glob import glob
from optparser import save_config_file
from auto_mutation import run_auto_mutation

__all__ = ["Job"]
_name = "runJob"


'''
Note to future self:
The Job atributes are kinda global variables here they  

Pdb files fate during simulation is as follows:
original_input.pdb is kept during all the simulation and then renamed as input.pdb 
input.pdb is the foldx result at each simulation stage, and is the "current" file
folded.pdb is a product of calculations usually soon renamed to input.pdb
output.pdb is the final product of the simulation 
Whole thing is a mess mostly due to how foldx works (which has now improved but this program is still behind on it)
'''


class Job:
    def __init__(self, config):
        for argName, argValue in config.items():
            setattr(self, argName, argValue)
            logger.debug(module_name=_name, msg="Setting %s to %s" % (argName, argValue))
        self.config = config
        if exists(self.work_dir):
            if not isdir(self.work_dir):
                raise logger.AggrescanError('Selected working directory: %s already exists and is not a directory. '
                                            'Quitting.' % self.work_dir,
                                            module_name=_name,
                                            work_dir_error=True)
            if self.overwrite and isfile(join(self.work_dir, "output.pdb")):
                raise logger.AggrescanError("The --overwrite options was seen. "
                                            "\nStopping the program to avoid overwriting files "
                                            "(workdir exists and contains output.pdb).",
                                            module_name=_name,
                                            work_dir_error=True)
            else:
                logger.warning(module_name=_name,
                               msg='Working directory already exists (possibly overwriting previous results -ow '
                                   'to prevent this behavior)')
        else:
            try:
                os.makedirs(self.work_dir)
            except OSError:
                raise logger.AggrescanError("Could not create working directory at %s" % self.work_dir,
                                            module_name=_name,
                                            work_dir_error=True)
        try:
            os.mkdir(self.tmp_dir)
        except OSError:
            pass
        if self.foldx:
            self.foldx_handler = fold(foldx_dir=self.foldx, work_dir=self.work_dir,
                                      script_dir=resource_filename("aggrescan", "data"),
                                      skip_minimization=self.subprocess)
        save_config_file(config=config, work_dir=self.work_dir)

    def run_job(self):
        logger.info(module_name=_name, msg='Starting aggrescan3d job on: %s with %s chain(s) selected' %
                                           (self.protein, self.chain or "all"))
        logger.info(module_name=_name, msg="Creating pdb object from: %s" % self.protein)
        pdbObj = pdb.Pdb(self.protein, output=join(self.work_dir, 'input.pdb'), chain=self.chain)
        pdbObj.validate()
        pdbObj.savePdbFile(path=join(self.work_dir, "original_input.pdb"))
        pdbObj.savePdbFile(path=join(self.tmp_dir, "input.pdb"))
        if self.mutate:
            if self.foldx:
                mutation = self.find_mutations(pdb_obj=pdbObj)
                self.foldx_handler.build_mutant(working_dir=self.tmp_dir, mutation_list=mutation)
            else:
                raise logger.AggrescanError("FoldX required for mutation analysis. To run aggrescan on a mutant without"
                                            " FoldX provide a mutant pdb file and run Aggrescan3D on it.",
                                            module_name=_name)

        if self.foldx:
            self.foldx_handler.minimize_energy(working_dir=self.tmp_dir)
        else:
            logger.info(module_name=_name,
                        msg="FoldX not utilized. Treating input pdb file as it was already optimized.")
            pdbObj.savePdbFile(path=join(self.tmp_dir, "folded.pdb"))

        if self.dynamic:
            os.chdir(self.work_dir)
            shutil.move(join(self.tmp_dir, "folded.pdb"), "input.pdb")
            run_cabs(config=self.config)
            shutil.move(join(self.work_dir, "folded.pdb"), join(self.tmp_dir, "folded.pdb"))
        analyze(config=self.config, target="folded.pdb", working_dir=self.tmp_dir, agg_work_dir=self.work_dir)
        if self.movie:
            self.create_movie()
        self.post_processing(work_dir=self.tmp_dir, final=True, model_name="folded")
        self.cleanup()
        if self.auto_mutation:
            self.check_auto_mut(pdb_obj=pdbObj)     # This will also modify the auto_mutation excluded part slightly
            run_auto_mutation(work_dir=self.work_dir, options=self.auto_mutation,
                              foldx_loc=self.foldx, distance=self.distance)

    def post_processing(self, work_dir="", final=True, model_name="folded"):    # TODO remove this function?
        prepare_output(work_dir=work_dir, final=final, model_name=model_name)

    def find_mutations(self, pdb_obj=None):
        chain_numbering = pdb_obj.getChainIdxResname()
        t = json.loads(chain_numbering)
        to_mutate = []
        for row in self.mutate:
            oi = str(row['idx'])
            on = str(row['oldres'])
            nn = str(row['newres'])
            ch = str(row['chain'])
            found = False
            try:
                for r in t[ch]:
                    if r['resname'] == on and r['chain'] == ch \
                            and r['residx'] == oi:
                        to_mutate.append(on + ch + oi + nn)
                        found = True
                        break
            except KeyError:
                logger.warning(module_name=_name, msg="Mutation %s likely tried to mutate a chain "
                                                      "that doesn't exist." % json.dumps(row))
                logger.info(module_name=_name, msg="Available chains: %s" % t.keys())

            if not found:
                logger.warning(module_name=_name, msg="Could not find the requested mutation: %s" % json.dumps(row))
        if len(to_mutate) == 0:
            raise logger.AggrescanError("Mutations table provided but its parsing failed. "
                                        "Most likely all the provided mutations were incorrect "
                                        "(referring to non existing residues, numbering errors, etc.)",
                                        module_name=_name)
        mutation = ",".join(to_mutate).strip() + ";"
        logger.debug(module_name=_name, msg="Mutation list: %s" % mutation)
        return mutation

    def check_auto_mut(self, pdb_obj=None):
        print self.auto_mutation
        if len(self.auto_mutation) > 2:
            chain_numbering = pdb_obj.getChainIdxResname()
            t = json.loads(chain_numbering)
            counted = 0
            for row in self.auto_mutation[2]:
                oi = str(row['idx'])
                ch = str(row['chain'])
                found = False
                try:
                    for r in t[ch]:
                        if r['chain'] == ch \
                                and r['residx'] == oi:
                            found = True
                            counted += 1
                            break
                except KeyError:
                    logger.warning(module_name="Auto_mut", msg="Attempted to exclude a residue that probably"
                                                               "doesn't exist. (%s)" % json.dumps(row))
                    logger.info(module_name="Auto_mut", msg="Available chains: %s" % t.keys())

                if not found:
                    logger.warning(module_name="Auto_mut", msg="Couldn't find the residue number %s in chain %s to exclude from "
                                                               "auto mutation" % (oi, ch))
            if counted == 0:
                logger.critical(msg="Residues to exclude from automated mutations provided but none "
                                            "could be found in the pdb file.",
                                            module_name="Auto_mut")
            # Parse it into "mutation syntax" for easier comparisons later on
            self.auto_mutation = list(self.auto_mutation)
            self.auto_mutation[2] = [i['idx'] + i["chain"] for i in self.auto_mutation[2]]

    def cleanup(self):
        """Move output from the temporary directory to work directory before the former is deleted"""
        os.chdir(self.tmp_dir)
        shutil.move(join(self.tmp_dir, "A3D.csv"), join(self.work_dir, "A3D.csv"))
        shutil.move(join(self.tmp_dir, "output.pdb"), join(self.work_dir, "output.pdb"))
        shutil.move(join(self.work_dir, "original_input.pdb"), join(self.work_dir, "input.pdb"))
        extensions = [".svg", ".png"]
        for ext in extensions:
            for f in glob("*%s" % ext):
                shutil.move(f, join(self.work_dir, f))

    def create_movie(self):
        """
        First uses paintit.py to create movie frame png, then run avconv to create a movie
        This is a legacy function while it should work it's no longer really relevant or updated/tested
        """
        os.chdir(self.tmp_dir)
        pymCmd = "pymol -qc %s -- s input.pdb input_output ." % resource_filename("aggrescan", join("data", "paintit.py"))
        logger.debug(module_name="pyMol", msg="Pymol ran with: %s" % pymCmd)

        try:
            out,err = Popen(pymCmd,stdout=PIPE,stderr=PIPE,shell=True).communicate()
            if err:
                logger.critical(module_name="pyMol",
                                        msg="Pymol encountered an error: %s Movie creation failed." % err.strip("\n"))
                return
        except OSError as e:
            logger.debug(module_name="pyMol", msg="Exception caught: %s" % e)
            logger.critical(module_name="pyMol", msg="OSError while launching pymol. Perhaps it's not installed?")
            return

        self.movie = self.movie.strip()
        if self.movie == "mp4":
            av_cmd = 'avconv  -r 8 -i mov%05d.png  -vcodec libx264 -pix_fmt yuv420p -profile:v baseline ' \
                     '-preset slower -crf 18 -vf "scale=trunc(in_w/2)*2:trunc(in_h/2)*2" clip.mp4'
            logger.info(module_name="Movie",msg="Creting movie with %s format" % self.movie)
        elif self.movie == "webm":
            av_cmd = 'avconv  -r 8 -i mov%05d.png -c:v libvpx -c:a libvorbis -pix_fmt yuv420p ' \
                     '-b:v 2M -crf 5 -vf "scale=trunc(in_w/2)*2:trunc(in_h/2)*2" clip.webm'
            logger.info(module_name="Movie",msg="Creting movie with %s format" % self.movie)
        else:
            logger.warning(module_name="Movie",msg="Wrong movie format specified: %s Using webm instead." % self.movie)
            av_cmd = 'avconv  -r 8 -i mov%05d.png -c:v libvpx -c:a libvorbis -pix_fmt yuv420p ' \
                     '-b:v 2M -crf 5 -vf "scale=trunc(in_w/2)*2:trunc(in_h/2)*2" clip.webm'
            self.movie = "webm"

        try:
            logger.debug(module_name="Avconv", msg="Avconv ran with: %s" % av_cmd)
            out, err = Popen(av_cmd,stdout=PIPE, stderr=PIPE, shell=True).communicate()
            if err:
                logger.debug(module_name="Avconv",
                                    msg="Avconv output: %s" % err.strip("\n"))
            if self.movie == "mp4":
                shutil.move("clip.mp4", join(self.work_dir,"clip.mp4"))
            else:
                shutil.move("clip.webm", join(self.work_dir,"clip.webm"))

        except OSError as e:
            logger.critical(module_name="Avconv", msg="OSError while launching avconv. Perhaps it's not installed?")
            logger.debug(module_name="Avconv", msg="Exception caught: %s" % e)

        except IOError as e2:
            logger.critical(module_name="Avconv", msg="Couldn't move the clip to working directory. "
                                                      "Movie creation likely failed")
            logger.debug(module_name="Avconv", msg="Exception caught: %s" % e2)

        finally:
                file_list = os.listdir(self.tmp_dir)
                for f in file_list:
                    if ".png" in f:
                        os.remove(f)

    def get_tempdir(self):
        return self.tmp_dir



# -*- coding: utf-8 -*-

"""Module that handles argument parsing and stores default program options"""
import argparse
from os import getcwd
from os.path import abspath, join, isfile
import json
import re
import logger
from pkg_resources import resource_filename


_name = "AggParser"

_shorts = {
    'i': 'protein',
    'm': 'mutate',
    'd': 'dynamic',
    'D': 'distance',
    'w': 'work_dir',
    'O': 'overwrite',
    'v': 'verbose',
    'C': 'chain',
    'f': 'foldx',
    'M': 'movie',
    'n': 'naccess',
    'o': 'output',
    'r': 'remote',
    'cabs': "cabs_dir",
    'cabs_config': 'cabs_config',
    'n_models': 'n_models',
    'am': 'auto_mutation'
}


class ExtendAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        items = getattr(namespace, self.dest) or []
        items.extend(values)
        setattr(namespace, self.dest, items)


def parse(options):
    parser = get_parser_object()
    read = parser.parse_args(options)
    data, mut = [], []
    if read.config_file:
        data, mut = parse_config_file(read.config_file)
    data.extend(options)
    read = parser.parse_args(data)
    auto_mutation = _parse_auto_mut(read.auto_mutation)

    # This is a safety measure in case a path is changed inside the program and the relative would be no longer valid
    cabs_config, foldx, naccess = '', '', ''
    if read.cabs_config:
        cabs_config = abspath(read.cabs_config)
    if read.foldx != "argument_not_used" and read.foldx is not None:   # Prevent assiging a path to not-used argument
        read.foldx = abspath(read.foldx)
    if read.naccess:
        naccess = abspath(read.naccess)
    foldx = _handle_foldx(read.foldx)
    work_dir = abspath(read.work_dir)
    tmp_dir = join(work_dir, "tmp")
    new_mut = _parse_mut(read.mutate)

    if new_mut:
        mut.extend(new_mut)
    # Catch some of the unwanted option combinations here.
    if new_mut and auto_mutation:
        raise logger.InputError("Both manual and automated mutations selected. Please select just one of those as they are exclusive.")
    if auto_mutation and read.dynamic:
        raise logger.InputError("Automated mutations are not (yet?) supported with dynamic mode.")
    if auto_mutation and not foldx:
        raise logger.InputError("Automated mutations require FoldX to work. Please specify a path with -f.")

    options = {
        'protein': read.protein,
        'dynamic': read.dynamic,
        'mutate': mut,
        'distance': read.distance,
        'work_dir': work_dir,
        'overwrite': read.overwrite,
        'verbose': read.verbose,
        'chain': read.chain,
        'movie': read.movie,
        'naccess': naccess,
        "foldx": foldx,
        'auto_mutation': auto_mutation,
        'remote': read.remote,
        "cabs_dir": read.cabs_dir,
        "cabs_config": cabs_config,
        "n_models": read.n_models,
        "tmp_dir": tmp_dir,
        "subprocess": read.subprocess
    }
    return options


def parse_config_file(filename):
    parsed = []
    option_pattern = re.compile(
        r'(?P<option>[^:=]*)'
        r'[:=]'
        r'(?P<value>.*)$'
    )
    mutation_pattern = re.compile(r'{+.+\}+')
    path = abspath(filename)
    logger.debug(module_name=_name, msg="Parsing data from: %s" % path)
    to_mutate = []
    try:
        with open(path, 'r') as f:
            for line in f:
                if line == '' or line[0] in ';#\n':
                    continue
                elif mutation_pattern.search(line):
                    try:
                        mutation = mutation_pattern.findall(line)[0]
                        mutation = re.sub("\'", "\"", mutation)
                        to_mutate.append(json.loads(mutation))
                    except Exception:
                        logger.critical(module_name=_name,
                                        msg="Error while loading the mutation table in line:\n %s" % line.strip("\n"))
                        raise logger.InputError("Failed to properly load config file. Quitting.")
                elif line.startswith("am") or line.startswith("auto_mutation"): # Special case, this is getting super messy
                    parsed.append("--auto_mutation")
                    parsed.append(" ".join(line.split()[1:]))
                else:
                    match = option_pattern.match(line)
                    if match:
                        option, value = match.groups()
                        option = option.strip()
                        if option in _shorts.values():
                            if value.strip() != "True":
                                parsed.append("--"+option)
                                parsed.append(value.strip())
                            else:
                                parsed.append("--"+option)
                        elif option in _shorts.keys():
                            if value.strip() != "True":
                                parsed.append("-"+option)
                                parsed.append(value.strip())
                            else:
                                parsed.append("-"+option)

                        else:
                            logger.warning(module_name=_name,
                                           msg="option: %s seen in config file not recognized" % option)
                    elif line.strip() in _shorts.values():
                        parsed.append("--"+line.strip())
                    elif line.strip() in _shorts.keys():
                        parsed.append("-"+line.strip())
                    else:
                        logger.warning(module_name=_name, msg="line: %s couldn't be parsed" % line.strip("\n"))

    except IOError:
        raise logger.InputError("Could not load the file %s" % path)
    return parsed, to_mutate


def save_config_file(config, work_dir):
    tmp = ""
    for argName, argValue in config.items():
        if type(argValue) == list:      # This will catch mutations
            for mutation in argValue:
                tmp += json.dumps(mutation) + "\n"  # assuming its a mutation entry, which should be dumps'able
        elif type(argValue) == tuple:   # This will catch automated mutations, as long as no new tuples/lists are introduced this works
            if len(argValue) > 2:
                tmp += "%s: %d %d %s\n" % (argName, argValue[0], argValue[1], " ".join([i['idx']+i['chain'] for i in argValue[2]]))
            else:
                tmp += "%s: %d %d\n" % (argName, argValue[0], argValue[1])
        elif argValue and (argName in _shorts.keys() or argName in _shorts.values()):
            tmp += "%s : %s\n" % (argName, argValue)

    logger.to_file(filename=join(work_dir, "config.ini"), content=tmp)


def get_parser_object():
    parser = argparse.ArgumentParser(
        description='Aggrescan 3D standalone application. Please read and run configReadme.ini to test if the program works.')
    parser.register('action', 'extend', ExtendAction)

    parser.add_argument('-c',
                        '--config_file',
                        metavar='',
                        type=str,
                        help='''Input file with specified config options. One can specify a config file and some additional flags
                             Command line options have priority over config file. 
                             For further help on creating config files please refer to configReadme.ini''')

    parser.add_argument('-i',
                        '--protein',
                        metavar='',
                        type=str,
                        help='Input pdb code or a path to file on which the simulation is performed')

    parser.add_argument('-d',
                        '--dynamic',
                        default=False,
                        action='store_true',
                        help='Use the dynamic module for the simulation. '
                             'CABS flex will be used to perform a protein dynamics simulation '
                             'If CABS is not installed but present in the filesystem use --cabs_dir to provide a path to its location '
                             '--cabs_config can be used to provide a CABS config file '
                             '--n_models can be used to decide how many models will be taken into further calculations')

    parser.add_argument('-m',
                        '--mutate',
                        metavar='',
                        default='',
                        type=str,
                        nargs='+',
                        action='extend',
                        help="""Provide a residue mutation. This option can be used multiple times to provide multiple mutations.
                                 The format is: <Old residue><New residue><Residue number><Chain ID> 
                                 For example: MW1A to change the first residue of chain A from methionine to tyrosine.
                                 Please note that this option requires FoldX usage.""")

    parser.add_argument('-D',
                        '--distance',
                        metavar='',
                        default=10,
                        type=int,
                        help='Distance cutoff for naccess calculations. Default: 10')

    parser.add_argument('-w',
                        '--work_dir',
                        metavar='',
                        default=getcwd(),
                        type=str,
                        help="Sets the directory in which the simulation is performed")

    parser.add_argument('--cabs_dir',
                        metavar='',
                        default='',
                        type=str,
                        help="If used CABS will be run from the provided directory. Should be used when CABS is not installed"
                             "or one wants to use a specific CABS version for the dynamic simulations")

    parser.add_argument('--cabs_config',
                        metavar='',
                        default='',
                        type=str,
                        help="If used CABS will be run with selected config file. "
                             "Aggrescan will not parse the config in any way so its up to the user to supply a working config file."
                             "For debugging purposes use verboisty 3 or higher, "
                             "CABS simulation log will be then kept, see the CABSlog file for details on what went wrong with the CABS run."
                             "Few caveats include:"
                             "CABS work-dir should not be used"
                             "if --n_models is used it will overwrite --clustering-medoids in the config file specified by this option"
                             "CABS input option will be overwritten"
                             "--aa-rebuild and --remote will be specified by default")

    parser.add_argument('--n_models',
                        metavar='',
                        default=12,
                        type=int,
                        help='Sets the number of models CABS will generate in the dynamic analysis.'
                             'This option will overwrite cabs config file if you chose both options'
                             ' '
                        )

    parser.add_argument('-r',
                        '--remote',
                        default=False,
                        action="store_true",
                        help='Automatically redirects output to a Aggrescan.log file created in the working directory,'
                             'turns off log coloring. Piping standard error will not work with this option.'
                             'If a log file already exists it will be appended to.')

    parser.add_argument('-O',
                        '--overwrite',
                        default=False,
                        action='store_true',
                        help="""If the working directory already exists and contains an output.pdb file 
                         this will stop the program to prevent overwriting possible results that might be there""")

    parser.add_argument('-f',
                        '--foldx',
                        metavar='',
                        type=str,
                        nargs='?',
                        default="argument_not_used",
                        help="""The program will attempt to run foldX on the provided pdb file or code 
                                 using the path provided in this option. Once the path is provided subsequent runs can 
                                 can omit the path and simply use -f (it will be written in program data""")

    parser.add_argument('-v',
                        '--verbose',
                        metavar='',
                        default=2,
                        type=int,
                        help="""Sets the verbosity of the program 
                                 0 for silent mode (only critical messages) and 4 for maximum verbosity. 
                                 Verbosity level 3 might raise some Exceptions that would otherwise only be logged as critical messages""")

    parser.add_argument('-C',
                        '--chain',
                        metavar='',
                        default='',
                        type=str,
                        help="""Only consider a specific chain while performing calculations""")

    parser.add_argument('-M',
                        '--movie',
                        metavar='',
                        default='',
                        type=str,
                        help="""Create a short movie of the result protein. Available formats: mp4, webm
                                 Usage: -m mp4
                                 Please note pymol and avconv are used and need to be installed for this option to work""")
    parser.add_argument('-n',
                        '--naccess',
                        metavar="",
                        default="",
                        type=str,
                        help="""Run the simulation with naccess instead of freeSasa. 
                                Run this argument with a path to the naccess program exectuable.""")

    parser.add_argument('-am',
                        '--auto_mutation',
                        metavar='',
                        default='is_not_used',  # The actual option is in _parse_auto_mut, this is to recognize the argument was NOT used
                        nargs='?',
                        type=str,
                        help="Automatically mutate most aggregation prone residues The option should be formatted as follows: "
                             "number_of mutations number_of_used_processes [list of residues that are not to be mutated]"
                             " The list is optional and if no arguments are specified two resides using two threads will be attempted"
                              "The list should follow the naming convention of <Residue number><Chain ID>" 
                              "For example: '2 2 1A' to mutate 2 residues using 2 processes at once and exclude the first residue of A chain"
                              " from the mutation candidates. Which will result in 8 mutants (4*2)")

    parser.add_argument('--subprocess',
                        default=False,
                        action="store_true",
                        help="""Internal option but can be used by the user - this is used to signal that the run is 
                                actually a subprocess of A3D's auto mutation procedure. Using this option means that no 
                                energy minimization will be perfomed before creating a mutant.""")

    return parser


def _parse_auto_mut(argument):
    _pattern = re.compile('(?P<id>[0-9]+)'
                          '(?P<chain>[A-Z])')

    if argument == "is_not_used":
        return None
    elif argument is None:  # Happens when option with no arguments is used
        return 2, 2
    else:
        argument = argument.strip()
        parsed_argument = argument.split()
        if parsed_argument[0] == ":" or parsed_argument[0] == "=":  # This can happen in a config file
            parsed_argument.pop(0)
        if len(parsed_argument) == 1:
            parsed_argument += "2"  # To add to the default
        try:
            n_mutations = int(parsed_argument[0])
        except ValueError:
            raise logger.InputError("Failed to convert the first argument for automated mutations to a number. Expected a number,"
                                    "got %s. Valid input would look like: -am '2 2 1A 2A'. Your input: %s" % (parsed_argument[0], argument))
        try:
            n_threads = int(parsed_argument[1])
        except ValueError:
            raise logger.InputError("Failed to convert the second argument for automated mutations to a number. Expected a number,"
                                    "got %s. Valid input would look like: -am '2 2 1A'. Your input: %s" % (parsed_argument[1], argument)
                                    )

        if len(parsed_argument) > 2:
            excluded_list = []
            for item in parsed_argument[2:]:
                try:
                    parse = _pattern.match(item).groups()
                    excluded_list.append({
                        'idx': parse[0],
                        'chain': parse[1]
                    })
                except AttributeError:
                    raise logger.InputError(
                        "Failed to parse excluded residues for aumated mutations. Residues should be provided  in a "
                        "<Residue ID><Chain> format i.e. 1A 2B etc. Failed item: %s. Whole input for the option: %s" % (item, argument))
            return n_mutations, n_threads, excluded_list
        else:
            return n_mutations, n_threads


def _parse_mut(mutation_list):
    p = re.compile('(?P<oldres>[A-Z])'
                   '(?P<newres>[A-Z])'
                   '(?P<id>[0-9]+)'
                   '(?P<chain>[A-Z])')
    out_list = []
    for mutation in mutation_list:
        try:
            parsed = p.match(mutation).groups()
            out_list.append({'oldres': parsed[0],
                             'newres': parsed[1],
                             'idx': parsed[2],
                             'chain': parsed[3]
                             })
        except AttributeError:
            raise logger.InputError(
                                "Your mutation %s was provided in wrong format. " % mutation +
                                "Correct formatting is <Oldres><Newres><ID><chain>. "
                                "For example MW1A. Mind the capital letters. ")
    return out_list


def _handle_foldx(user_input):
    """
    Either save the provdied FoldX path to a file or try to read the path from the file.
    If the user sepcified -f with no argument the user_input is None
    if the user never used the -f the user_input is default ("")
    """
    if user_input == "argument_not_used":    # The option was never selected
        return ""
    if user_input is not None:  # A path is actually provided
        with open(join(resource_filename("aggrescan", "data"), "foldx_dir.txt"), 'w') as f:
            f.write(user_input)
    else:
        if not isfile(join(resource_filename("aggrescan", "data"), "foldx_dir.txt")):
            raise logger.InputError("--foldx option used but no path has been specified. If you've never used the option "
                                    "you need to privde a path to FoldX directory. It can be ommited in subsequent runs "
                                    "as it will be saved in program's data")
        with open(join(resource_filename("aggrescan", "data"), "foldx_dir.txt"), 'r') as f:
            user_input = f.read().strip()
    return user_input



# -*- coding: utf-8 -*-

import os
import sys
from time import time, strftime, gmtime
import textwrap
import platform

_name = "Logger"

colors = {
    "blue": '\033[94m',
    "yellow": '\033[93m',
    'green': '\033[92m',
    'red': '\033[91m',
    'light_blue': '\033[96m',
    'purple': '\033[95m',
    'end': '\033[0m',
}

log_levels = {
    0: "[CRITICAL]",
    1: "[WARNING]",
    2: "[INFO]",
    3: "[OUT FILES]",
    4: "[DEBUG]"
}

color_prefix = {
    0: colors["red"] + "[CRITICAL]" + colors["end"],
    1: colors["yellow"] + "[WARNING]" + colors["end"],
    2: colors["blue"] + "[INFO]" + colors["end"],
    3: colors["purple"] + "[OUT FILES]" + colors["end"],
    4: colors["green"] + "[DEBUG]" + colors["end"],
}

_init_time = time()
_log_level = 2
_color = True
_stream = sys.stderr
_line_format = '%-20s %-19s%-75s %s\n'
_first_line_format = '%-20s %-19s%-75s \n'
_middle_line_format = '%-22s%-75s \n'
_last_line_format = '%-22s%-75s %s\n'
_line_break = 76
_remote = False
_prefix = color_prefix
_work_dir = ""


def setup(log_level=2, remote=False, work_dir=''):
    global _log_level, _color, _stream, _remote, _line_break, _prefix, _work_dir
    global _line_format, _middle_line_format, _first_line_format, _last_line_format
    _log_level = log_level
    _remote = remote
    _work_dir = work_dir
    if _remote or not sys.stderr.isatty():
        _color = False
        _line_format = '%-12s %-10s%-75s %s\n'
        _first_line_format = '%-12s %-10s%-75s \n'
        _middle_line_format = '%-22s %-75s \n'
        _last_line_format = '%-22s %-75s %s\n'
        _prefix = log_levels
    if _remote:
        _log_path = os.path.join(work_dir, "Aggrescan.log")
        try:
            _stream = open(_log_path, 'a+')
        except IOError:
            try:
                os.makedirs(work_dir)
                _stream = open(_log_path, 'a+')
            except OSError:
                warning(module_name=_name,
                        msg="Could not open a log file at %s. Writing to standard error instead." % _log_path)
            else:
                _stream.close()
                _stream = _log_path
        else:
            _stream.close()
            _stream = _log_path
    if platform.system() == "Windows":
        _color = False  # Windows doesn't support this way of coloring its logs in the command line
        _prefix = log_levels
    info(_name, 'Verbosity set to: ' + str(log_level) + " - " + log_levels[log_level])


def log_files():
    """
    :return: True if verbosity is high enough to save extra output (LOG FILE level)
    """
    return _log_level >= 3


def get_log_level():
    return _log_level


def coloring(color_name="light_blue", msg=""):
    if _color:
        return colors[color_name] + msg + colors["end"]
    return msg


def log(module_name="MISC", msg="Processing ", l_level=2, out=None):
    if out is None:
        # This is not great but should use stderr if there's a problem with log file, hopefully no race happens
        # This whole thing is attempting to force the log to really flush its data in real time
        try:
            out = open(_stream, 'a+')
        except TypeError:       # This should mean _stream is actually stderr
            out = _stream
    if l_level <= _log_level:
        t = gmtime(time() - _init_time)
        if len(msg) < _line_break:
            msg = _line_format % (
                _prefix[l_level], coloring(msg=module_name + ":", color_name='light_blue'),
                msg, strftime('(%H:%M:%S)', t)
            )
            out.write(msg)
            out.flush()
        else:
            lines = textwrap.wrap(msg, width=_line_break-1)
            first_line = _first_line_format % (
                _prefix[l_level], coloring(msg=module_name + ":", color_name='light_blue'), lines[0]
            )
            out.write(first_line)
            for lineNumber in xrange(1, len(lines) - 1):
                line = _middle_line_format % (" ", lines[lineNumber])
                out.write(line)
            final_line = _last_line_format % (" ", lines[-1], strftime('(%H:%M:%S)', t))
            out.write(final_line)
            out.flush()

        if out is not sys.stderr:
            out.close()


def critical(module_name="_name", msg=""):
    log(module_name=module_name, msg=msg, l_level=0)


def warning(module_name="_name", msg=""):
    log(module_name=module_name, msg=msg, l_level=1)


def info(module_name="_name", msg=""):
    log(module_name=module_name, msg=msg, l_level=2)


def log_file(module_name="_name", msg=""):
    log(module_name=module_name, msg=msg, l_level=3)


def debug(module_name="_name", msg=""):
    log(module_name=module_name, msg=msg, l_level=4)


def to_file(filename='', content='', msg='', allow_err=True, traceback=True):
    """

    :param filename: path for the file to be saved
    :param content: a string to be saved (be careful not to pass a string that is too long
    :param msg: optional: a message to be logged
    :param allow_err: if True a warning will be logged on OSError, if False program exit call will be made
    :param traceback: if True an Exception will be raised on exit call
    :return:
    """
    if filename and content:
        try:
            if os.path.isfile(filename):
                log_file(module_name=_name, msg="Overwriting %s" % filename)
            with open(filename, 'w') as f:
                f.write(content)
        except IOError:
            if allow_err:
                warning(module_name=_name, msg="IOError while writing to: %s" % filename)
            else:
                exit_program(module_name=_name, msg="IOError while writing to: %s" % filename, traceback=traceback)
    if msg:
        log_file(module_name=_name, msg=msg)


def exit_program(module_name=_name, msg="Shutting down", traceback=None, exc=None):
    """
    Exits the program, depending on options might do it quietly, raise, or print traceback
    :param module_name: string with  the calling module's name
    :param msg: string, message to be printed when the program exits
    :param traceback: str, if log level is high traceback will be printed
    :param exc: a specific exception passed by the caller
    :return: None
    """

    if exc:
        _msg = '%s: %s' % (msg, exc.message)
    else:
        _msg = msg
    critical(module_name=module_name, msg=_msg)
    if _log_level > 3 and traceback:
        _stream.write(traceback)
    sys.exit(1)


def record_exception(trace_stack):
    global _work_dir
    with open(os.path.join(_work_dir, "Aggrescan.error"), 'w') as f:
        file_header = "Aggrescan encountered an error and it wasn't one we were expecting. \n" \
                      "Please see the the following Traceback, perhaps it will be helpfull in understanding what happened.\n" \
                      "We would be grateful if you reported the incident to one of the authors so we can correct the error.\n"
        f.write(file_header)
        f.write(trace_stack)


class AggrescanError(Exception):
    """Generic class for Aggrescan Exceptions that should be handled or otherwise logged
        The subclasses are meant to do most of the work, so this is obviously work in progress
        and hopefully subclasses will do more in the future"""
    def __init__(self, *args, **kwargs):
        self.module_name = ""
        self.log_if_remote = ""
        self.work_dir_error = ""
        self.error_msg = ""
        self.err_file = ""
        if "module_name" in kwargs:
            self.module_name = kwargs["module_name"]
        if "err_file" in kwargs:    # Prevents the error log from being created - since the error is path/dir related
            self.err_file = kwargs["err_file"]
        if "work_dir_error" in kwargs:
            self.work_dir_error = True
        self.logger_msg = self.module_name + " encountered an error"
        if args:
            self.error_msg = args[0]
        Exception.__init__(self, *args)
    # This function and err_file serve a somewhat convoluted function - they split what is the message
    # to be displayed on screen and the one that would be logged in a file into two separate places using a proxy -
    # extra error file

    def generate_error_file(self):
        global _work_dir
        if not self.work_dir_error:
            out_file = os.path.join(_work_dir, "Aggrescan.error")
            file_header = "One of Aggrescan3D modules (%s) encountered an error. \n" % self.module_name
            with open(out_file, 'a+') as f:
                if self.err_file:
                    f.write(file_header)
                    with open(os.path.join(_work_dir, self.err_file), 'r') as err_file:
                        f.write(err_file.read())
                else:
                    f.write(file_header)
                    f.write(self.error_msg)


class FoldXError(AggrescanError):
    def __init__(self, *args, **kwargs):
        _module_name = "FoldX"
        AggrescanError.__init__(self, *args, module_name=_module_name, **kwargs)


class CabsError(AggrescanError):
    def __init__(self, *args, **kwargs):
        _module_name = "CABS"
        AggrescanError.__init__(self, *args, module_name=_module_name, **kwargs)


class ASAError(AggrescanError):
    def __init__(self, program_name, filename):
        _module_name = "ASA"
        msg = "Aggrescan encountered an error while performing accessibility calculations with %s. " \
              "See the details in the Aggrescan.error file in your working directory." % program_name
        AggrescanError.__init__(self, msg, module_name=_module_name, err_file=filename)


class InputError(AggrescanError):
    def __init__(self, *args):
        _module_name = "OptParser"
        AggrescanError.__init__(self, *args, module_name=_module_name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sqlite3
import os
from wtforms.validators import ValidationError
from flask import Response, request, flash, redirect, url_for
from a3d_gui import app
from utils import query_db, unique_id
from job_handling import delete_files, check_jobs
from create_config import reverse_config
from os.path import join, abspath, isfile
from parsePDB import PdbParser
"""
Module intended for simple database operations for the Aggrescan3D standalone app
CAUTION! Some of the functions here perform file deletion as well as process killing                        
Should you chose to modify basic settings or job handling and db managing functions do so with care 
"""


def create_new_database(new_database):
    """Create an empty database that follows this apps schema"""
    try:
        conn = sqlite3.connect(new_database)
        c = conn.cursor()
        c.execute('''CREATE TABLE user_queue (id INTEGER PRIMARY KEY, jid TEXT,
                     status TEXT DEFAULT 'pending', project_name TEXT, working_dir TEXT,
                     started DATETIME DEFAULT CURRENT_TIMESTAMP, pid INTEGER)''')
        c.execute('''CREATE TABLE project_details (jid TEXT, chains TEXT,
                     chain_sequence TEXT, chain_numbering TEXT, dynamic INTEGER, mutate INTEGER, 
                     distance INTEGER, mutation TEXT, foldx INTEGER, mutt_energy_diff REAL,
                     auto_mutation TEXT)''')
        c.execute('''CREATE TABLE pictures (jid TEXT, filename TEXT, type TEXT, model TEXT)''')
        c.execute('''CREATE TABLE system_data (id INTEGER PRIMARY KEY, foldx_path TEXT)''')
        foldx_loc = raw_input("Please provide a location for your FoldX installation if you have one "
                              "(or just press Enter this can be done later on)\n")
        conn.commit()
        # If the db didn't exist maybe the user wants to input the foldx location?
        try:
            if os.path.isdir(foldx_loc):
                if os.path.isfile(os.path.join(foldx_loc, 'rotabase.txt')):
                    c.execute("INSERT INTO system_data(foldx_path) VALUES(?)", [foldx_loc])
                else:
                    c.execute("INSERT INTO system_data(foldx_path) VALUES(?)", ["Not specified"])
                    print "%s Is not a valid FoldX directory (missing rotabase.txt file). " \
                          "Use the gui to specify a valid location" % foldx_loc
            else:
                c.execute("INSERT INTO system_data(foldx_path) VALUES(?)", ["/home/FoldX"])  # Default value for Docker Image usage
                print "%s is not a directory. If you wish to use foldX in the " \
                      "gui specify a valid address when starting the job" % foldx_loc
        except EOFError:
            c.execute("INSERT INTO system_data(foldx_path) VALUES(?)",
                      ["/home/FoldX"])  # Default value for Docker Image usage
        finally:
            conn.commit()
            conn.close()
    except sqlite3.Error as e:
        raise ValidationError("An error occurred while trying to create a database: %s" % e)
    finally:
        conn.close()


@app.route('/_delete_job/<jid>', methods=['GET'])
def delete_job(jid):
    """Delete a job from current database"""
    try:
        query_db("DELETE FROM user_queue WHERE jid=?", [jid], insert=True)
        query_db("DELETE FROM project_details WHERE jid=?", [jid], insert=True)
        query_db("DELETE FROM pictures WHERE jid=?", [jid], insert=True)
        flash("Job %s deleted." % jid, "info")
        return url_for('index_page')
    except sqlite3.Error as e:
        return Response("The database couldn't process this request. Error: %s" % str(e),
                        status=400, content_type="text/plain")




@app.route('/_hard_delete_job/<jid>', methods=['GET'])
def hard_delete_job(jid):
    """Delete a job from current database as well as the files from the computer"""
    try:
        folder_gone = delete_files(jid)
        query_db("DELETE FROM user_queue WHERE jid=?", [jid], insert=True)
        query_db("DELETE FROM project_details WHERE jid=?", [jid], insert=True)
        query_db("DELETE FROM pictures WHERE jid=?", [jid], insert=True)
        if folder_gone:
            flash("Job %s deleted. All files and the working directory were deleted." % jid, "info")
        else:
            flash("Job %s deleted. Aggrescan files were most likely deleted"
                  " but the directory still contains files." % jid, "info")
        return url_for('index_page')
    except sqlite3.Error as e:
        return Response("The database couldn't process this request. Error: %s" % str(e),
                        status=400, content_type="text/plain")


@app.route('/_add_job/', methods=['POST'])
def add_job_to_db():
    """Using a config file add a job to active database"""
    config_file = request.files['file']
    for line in config_file.stream.readlines():
        if "work_dir :" in line:
            project_work_dir = line.split(":")[1].strip()
    try:
        options, mutation = reverse_config(join(project_work_dir, "config.ini"))    # relies on A3D automatic saving of parsed config
    except UnboundLocalError:
        return Response("This file doesn't contain a working dir field which is necessary. "
                        "Please select a valid config.ini file.",
                        status=400, content_type="text/plain")
    jid = unique_id()
    if not isfile(join(project_work_dir, "input.pdb")) and not isfile(join(project_work_dir, "tmp", "input.pdb")):
        return Response("This project has probably just started and some important files (input.pdb) "
                        "are not yet created. Or have been deleted since. If its the former try again in a second.",
                        status=400, content_type="text/plain")
    project_name = request.form['text'] or "Custom project"
    mutate = 1 if mutation else 0
    dynamic = 1 if options['dynamic'] else 0
    foldx = 1 if options['foldx'] else 0
    auto_mutation = options['auto_mutation']
    query_db("INSERT INTO user_queue (jid, project_name, working_dir, pid)\
            VALUES(?,?,?,?)",
             [jid, project_name, project_work_dir, -100], insert=True)  # -100 should never be a valid pid
    query_db("INSERT INTO project_details (jid, dynamic, mutate, distance, foldx, auto_mutation) \
             VALUES(?,?,?,?,?,?)",
             [jid, dynamic, mutate, options['distance'],  foldx, auto_mutation], insert=True)
    try:
        with open(get_filepath(jid, "input.pdb"), 'r') as f:
            p = PdbParser(f, options['chain'])
    except IOError:
        with open(join(project_work_dir, "tmp", "input.pdb"), 'r') as f:
            p = PdbParser(f, options['chain'])
    input_seq = p.getSequence()
    chain_numbering = p.getChainIdxResname()
    if not options['chain']:
        chains = p.getChains()
    query_db("UPDATE project_details SET chains=?, chain_sequence=?, chain_numbering=? WHERE jid=?",
             ["".join(chains), input_seq, chain_numbering, jid], insert=True)

    check_jobs(jid)
    return url_for('job_status', jid=jid)


@app.route('/change_foldx_path', methods=['POST'])
def change_foldx_path():
    new_path = request.form['text']
    if not new_path:
        return Response("Empty path provided, not taking any action.", status=400, content_type="text/plain")
    if not isfile(join(new_path, "rotabase.txt")):
        return Response("%s Is not a valid FoldX directory (missing rotabase.txt file)." % new_path,
                        status=400, content_type="text/plain")
    query_db("UPDATE system_data SET foldx_path=? WHERE id=?", [new_path, 1], insert=True)
    return Response(new_path, status=200, content_type="text/plain")


@app.route('/_change_status/<jid>', methods=['POST'])
def change_job_status(jid):
    """Manually update a job status in case user wishes to do so"""
    new_status = request.data
    try:
        query_db("UPDATE user_queue SET status=? WHERE jid=?", [new_status, jid], insert=True)
        return Response("Query OK. Status changed to %s" % new_status, status=200, content_type="text/plain")
    except sqlite3.Error as e:
        return Response("Query failed! Error: %s" % str(e), status=400, content_type="text/plain")


def get_filepath(jid, filename):
    data_dir = query_db("SELECT working_dir FROM user_queue WHERE \
            jid=?", [jid], one=True)[0]
    return join(data_dir, filename)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import gzip
import uuid
import re
import sqlite3
from flask import g
from a3d_gui import app
from collections import OrderedDict
from os.path import join


def gunzip(filename):
    gunzipped = ".".join(filename.split(".")[:-1])
    with open(gunzipped, 'w') as f_out:
        with gzip.open(filename, 'rb') as f_in:
            f_out.writelines(f_in)


def unique_id():
    return hex(uuid.uuid4().time)[2:-1]


def regexp(expr, item):
    r = re.compile(expr)
    return r.match(item) is not None


def connect_db():
    """Connects to a specific database."""
    try:
        rv = sqlite3.connect(app.config['DATABASE'])
        rv.create_function('regexp', 2, regexp)
        rv.row_factory = sqlite3.Row
    except sqlite3.Error:
        raise
    return rv


def get_db():
    """
    Opens a new database connection if there is none yet for the
    current application context.
    """
    if not hasattr(g, 'sqlite_db'):
        g.sqlite_db = connect_db()
    return g.sqlite_db


def query_db(query, args=(), one=False, insert=False):
    mydb = get_db()
    cur = mydb.cursor()
    cur.execute(query, args)
    if insert:
        mydb.commit()
        cur.close()
        return cur.lastrowid
    else:
        rv = cur.fetchall()
        cur.close()
        return (rv[0] if rv else None) if one else rv


def status_color(status, shorter=True):
    if not status:
        return '<span class="label label-danger">\
                <i class="fa fa-exclamation-circle"></i> error</span>'
    if not shorter:
        if status == 'pending':
            return '<span class="label label-warning">\
                    <i class="fa fa-spin fa-spinner"></i> pending \
                    </span>'
        elif status == 'missing_files':
            return '<span class="label label-info">\
                    <i class="fa fa-sliders"></i> Missing files</span>'
        elif status == 'running':
            return '<span class="label label-primary">\
                    <i class="fa fa-bolt"></i> running</span>'
        elif status == 'error':
            return '<span class="label label-danger">\
                    <i class="fa fa-exclamation-circle"></i> error</span>'
        elif status == 'done':
            return '<span class="label label-success">\
                    <i class="fa fa-check-circle-o"></i> done</span>'
    else:
        if status == 'pending':
            return '<span class="label label-warning">\
                    <i class="fa fa-spin fa-spinner"></i> pending</span>'
        elif status == 'missing_files':
            return '<span class="label label-info">\
                    <i class="fa fa-sliders"></i> Missing files</span>'
        elif status == 'running':
            return '<span class="label label-primary">\
                    <i class="fa fa-bolt"></i> running</span>'
        elif status == 'error':
            return '<span class="label label-danger">\
                    <i class="fa fa-exclamation-circle"></i> error</span>'
        elif status == 'done':
            return '<span class="label label-success">\
                    <i class="fa fa-check-circle-o"></i> done</span>'


def sort_dict(my_dict):
    new_dict = OrderedDict()
    for key, value in sorted(my_dict.iteritems(), key=lambda (k, v): (v, k), reverse=True):
        new_dict[key] = value
    return new_dict


def aminoacids():
    aa_dict = {'A': 'ALA', 'R': 'ARG', 'N': 'ASN', 'D': 'ASP', 'C': 'CYS',
               'E': 'GLU', 'Q': 'GLN', 'G': 'GLY', 'H': 'HIS', 'I': 'ILE',
               'L': 'LEU', 'K': 'LYS', 'M': 'MET', 'F': 'PHE', 'P': 'PRO',
               'S': 'SER', 'T': 'THR', 'W': 'TRP', 'Y': 'TYR', 'V': 'VAL',
               'X': 'UNK'}
    aa_dict_F = {'A': 'alanine', 'R': 'arginine', 'N': 'asparagine',
                 'D': 'aspartic acid', 'C': 'cysteine', 'E': 'glutamic acid',
                 'Q': 'glutamine', 'G': 'glycine', 'H': 'histidine',
                 'I': 'isoleucine', 'L': 'leucine', 'K': 'lysine',
                 'M': 'methionine', 'F': 'phenylalanine', 'P': 'proline',
                 'S': 'serine', 'T': 'threonine', 'W': 'tryptophan',
                 'Y': 'tyrosine', 'V': 'valine', 'X': 'unknown'}
    aa_t = []
    for k in sorted(aa_dict, key=aa_dict.get):
        if k != 'X':
            aa_t.append((aa_dict[k], aa_dict_F[k], k))
    return aa_t


def get_undefined_job_ids(q):
    to_check = []
    for i in q:
        if i['status'] == 'pending' or i['status'] == 'missing_files' or i['status'] == 'running':
            to_check.append(i['jid'])
    return to_check


def parse_out(q):
    out = []
    for row in q:
        if row['datet']:
            dtt = row['datet']
        else:
            dtt = "-"
        l = {'project_name': row['project_name'], 'jid': row['jid'],
             'date': dtt, 'status': status_color(row['status'])}

        out.append(l)
    return out





#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import math
import csv
import logger

# minimum and maximum surface exposition to be considered
min_surf = 10
max_surf = 55
_name = "agg3D"


class Residue:
    def __init__(self, chain, resi, resn, coords, rsa, score_matrix):
        if chain == ' ':
            self.chain = '-'
        else:
            self.chain = chain
        self.resi = resi
        self.resn = get_one_letter(resn)   # we store it in one-letter code
        self.coords = coords
        self.score_matrix = score_matrix
        self.agg_sup = self.calc_res_agg_fin(rsa=rsa)
        self.agg_fin = 0

    def __str__(self):
        return self.chain+self.resn

    def set_agg_fin(self, agg_dis_sum):
        self.agg_fin = self.agg_sup + agg_dis_sum

    def id(self):
        return self.chain+self.resn+self.resi

    def set_rsa(self, rsa):
        self.agg_sup = self.calc_res_agg_fin(rsa=rsa)

    def calc_res_agg_fin(self, rsa):
        if rsa < min_surf:
            return 0
        else:
            if rsa > max_surf:
                rsa = max_surf
            return float(self.score_matrix[self.resn]) * 0.0599 * math.exp(0.0521 * rsa)


class Protein:
    def __init__(self, pdb_code, residues, out_dir, max_dist):
        self.name = pdb_code
        self.residues = residues
        self.out_dir = out_dir
        self.max_dist = max_dist

    def __str__(self):
        return self.name

    def add_residue(self, res):
        self.residues.append(res)

    def change_res_rsa(self, res_number, res_id, rsa):
        if self.residues[res_number].id() == res_id:
            self.residues[res_number].set_rsa(rsa)
        else:
            logger.critical(module_name=_name,
                            msg="Error while parsing freeSasa output. Probably freeSasa failed quietly.")
            logger.critical(module_name=_name,
                            msg="Failed to assign rsa to %s proper id is: %s " %
                                (res_id, self.residues[res_number].id()) +
                            "(ChainID, residue's one letter code, residue's ID)")
            raise logger.AggrescanError("Failed to parse freeSasa output. Qutting.",
                                        module_name=_name)

    def calc_agg_fin(self):
        """Calculate the agg_dis_sum for every Residue in the Protein"""
        dist_correction = -2.56 / self.max_dist
        for central in self.residues:
            agg_dis_sum = 0
            if not central.agg_sup == 0:
                for peripheric in self.residues:
                    dist = get_distance(central.coords, peripheric.coords)
                    if dist < self.max_dist and not dist == 0:
                        agg_dis_sum += peripheric.agg_sup * 1.2915 * math.exp(dist_correction*dist)
            central.set_agg_fin(agg_dis_sum)

    def short_summary(self):
        """Return a short summarized output and identify it with a leading #"""
        total_sum = 0
        maximum = 0
        minimum = 0
        pos_auc = 0
        neg_auc = 0
        prev_score = 0
        res_number = len(self.residues)
        for res in self.residues:
            total_sum += res.agg_fin
            if res.agg_fin > maximum:
                maximum = res.agg_fin
            if res.agg_fin < minimum:
                minimum = res.agg_fin
            if res.agg_fin != 0:
                if prev_score != 0:
                    auc = (res.agg_fin + prev_score)/2
                    if auc > 0:
                        pos_auc += abs(auc)
                    else:
                        neg_auc += abs(auc)
                prev_score = res.agg_fin
        return '#' + self.name + " " + ' '.join(str("%.4f" % round(val, 4))
                                                for val in (total_sum,
                                                total_sum/res_number,
                                                maximum,
                                                minimum,
                                                pos_auc))

    def long_output(self):
        """Return the complete output and identify it with a leading //"""
        output = ""
        for res in self.residues:
            output += '//' + " ".join((res.chain, res.resi, res.resn)) + " %.4f\n" % round(res.agg_fin, 4)
        return output

    def out_csv(self):
        """Save the complete output in a csv file"""
        csv_file = os.path.join(self.out_dir, "A3D.csv")
        c = csv.writer(open(csv_file, "w"))
        if os.stat(csv_file).st_size == 0:
            c.writerow(["protein", "chain", "residue", "residue_name", "score"])
        for res in self.residues:
            c.writerow([self.name.split("/")[-1], res.chain, res.resi, res.resn, "%.4f" % round(res.agg_fin, 4)])

    def get_residues(self):
        return self.residues


def get_one_letter(raw):
    """Return the one letter code for a residue"""
    if len(raw) > 3:
        aa = raw[-3:]
    else:
        aa = raw

    three_letters = aa.capitalize()

    conversion = {"Ala": "A", "Arg": "R", "Asn": "N", "Asp": "D", "Cys": "C",
                  "Glu": "E", "Gln": "Q", "Gly": "G", "His": "H", "Ile": "I",
                  "Leu": "L", "Lys": "K", "Met": "M", "Phe": "F", "Pro": "P",
                  "Ser": "S", "Thr": "T", "Trp": "W", "Tyr": "Y", "Val": "V"}
    try:
        one_letter = conversion[three_letters]
    except KeyError:
        logger.warning(module_name=_name,
                       msg='Could not recognize the following residue symbol: "%s"' % raw)
        one_letter = None
    return one_letter


def get_distance(a, b):
    """Return the distance between two cartesian tri dimensional points"""
    return math.sqrt((a[0] - b[0])**2 +
                     (a[1] - b[1])**2 +
                     (a[2] - b[2])**2)


def parse_matrix(filename=''):
    """Parse the matrix input into a dictionary"""
    with open(filename, 'r') as fh:
        matdict = {}
        for line in fh.readlines():
            if not line == '\n':
                pair = line.strip().split(" ")
                matdict[pair[0]] = pair[1]

    return matdict


def parse_naccess(pdb_code, work_dir, score_matrix, max_dist):
    """Parse the output of naccess into an object of class Protein"""
    with open(os.path.join(work_dir,pdb_code) + ".rsa", "r") as fh:
        rsa_dict = {line[3:8].strip() +  # resn
                    line[8] +   # chain
                    line[9:16].strip():  # resi
                    float(line[22:28].strip())  # rsa
                    for line in fh.readlines()
                    if line.startswith('RES')}

    my_protein = Protein(pdb_code=pdb_code, residues=[], out_dir=work_dir, max_dist=max_dist)
    center_atom = 'CA'  # atom to be considered the center of the residue

    with open(os.path.join(work_dir,pdb_code) + '.asa', 'r') as fh:
        for line in fh.readlines():
            # use data only from alpha carbons
            if line[12:16].strip() == center_atom:
                # extract the data according to the characteristics of the pdb format
                my_chain = line[21]
                my_resi = line[22:29].strip()
                my_resn = line[17:20].strip()
                x = float(line[30:38].strip())
                y = float(line[38:46].strip())
                z = float(line[46:54].strip())
                res_id = my_resn + my_chain + my_resi
                my_rsa = rsa_dict[res_id]
                my_protein.add_residue(Residue(chain=my_chain,
                                               resi=my_resi,
                                               resn=my_resn,
                                               coords=(x, y, z),
                                               rsa=my_rsa,
                                               score_matrix=score_matrix))
    return my_protein


def parse_freesasa(pdb_code, work_dir, filename, score_matrix, max_dist):
    my_protein = Protein(pdb_code=pdb_code, residues=[], out_dir=work_dir, max_dist=max_dist)
    residue_count = 0
    central_atom = 'CA'   # atom to be considered the center of the residue
    with open(os.path.join(work_dir, filename), 'r') as f:
        for line in f:
            if line.startswith("ATOM "):  # this is the same as naccess asa file
                if line[12:16].strip() == central_atom:
                    # extract the data according to the characteristics of the pdb format
                    my_chain = line[21]
                    my_resi = line[22:29].strip()
                    my_resn = line[17:20].strip()
                    x = float(line[30:38].strip())
                    y = float(line[38:46].strip())
                    z = float(line[46:54].strip())
                    my_protein.add_residue(Residue(chain=my_chain,
                                                   resi=my_resi,
                                                   resn=my_resn,
                                                   coords=(x, y, z),
                                                   rsa=0,
                                                   score_matrix=score_matrix))
            elif line.startswith("RES"):   # equivalent to naccess rsa file
                this_id = line[8] + get_one_letter(line[3:8].strip()) + line[9:16].strip()
                rsa = float(line[22:28].strip())
                my_protein.change_res_rsa(res_number=residue_count, res_id=this_id, rsa=rsa)
                residue_count += 1

    return my_protein


def run(pdb_file='', mat_file='', max_dist='', work_dir='', naccess=False):
    score_matrix = parse_matrix(filename=mat_file)
    pdb_code = pdb_file.split(".")[0]
    if naccess:
        my_protein = parse_naccess(pdb_code=pdb_code, work_dir=work_dir,
                                   score_matrix=score_matrix, max_dist=max_dist)
    else:
        my_protein = parse_freesasa(pdb_code=pdb_code, work_dir=work_dir,
                                    filename="sasa.out", score_matrix=score_matrix, max_dist=max_dist)
    my_protein.calc_agg_fin()
    my_protein.out_csv()
    short_summary = my_protein.short_summary()+"\n"
    long_summary = my_protein.long_output()
    return short_summary + long_summary



# -*- coding: utf-8 -*-

import logger
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
import numpy as np
import re
import os
import warnings
import json
from os.path import join
from collections import OrderedDict

_name = "postProc"


def params():
    """
    Sets some global parameters for the plots
    :return: None
    """
    plt.rcParams['axes.facecolor'] = 'f5f5f5'
    plt.rcParams['axes.edgecolor'] = '0.45'
    plt.rcParams['axes.axisbelow'] = True
    plt.rcParams['axes.labelcolor'] = '0.45'
    plt.rcParams['text.color'] = '0.45'
    plt.rcParams['xtick.color'] = '0.45'
    plt.rcParams['ytick.color'] = '0.45'
    plt.rcParams['xtick.major.pad'] = 4
    plt.rcParams['ytick.major.pad'] = 5
    plt.rcParams['xtick.major.width'] = 1
    plt.rcParams['ytick.major.width'] = 1


def get_scores(out_file):
    """

    :param out_file: path to a csv file generated by Aggrescan3D
    :return: dictionary - {chainID:[residue ID, agg3d score, residue label]},
             dictionary - {chainID:{"min_value":val,"max_value":val,
                            "total_value":val,"avg_value":val}}
    """

    pattern = re.compile(r"^(.*),(.*),(.*),(.*),(.*)$", re.M)
    with open(out_file, 'r') as f:
        data = pattern.findall(f.read().replace("\r", ""))[1:]

    chains = set([i[1] for i in data])
    chains.add("All")
    dat = {}
    scores = {}
    stats = {}
    resNumber = 1
    for chain_id in chains:
        dat[chain_id] = []
        scores[chain_id] = []
    for line in data:
        if len(line) != 5:
            continue

        chain = line[1]
        label = line[3] + line[2]  # One letter code + residue ID
        aggScore = float(line[4])
        scores[chain].append(aggScore)
        scores["All"].append(aggScore)
        if abs(aggScore) > 1e-10:  # Skip residues with 0 score
            dat[chain].append((resNumber, aggScore, label))
        resNumber += 1

    for chain in chains:
        min3d = min(scores[chain])
        max3d = max(scores[chain])
        sum3d = np.sum(scores[chain])
        avg3d = np.round(sum3d / len(scores[chain]), decimals=4)
        stats[chain] = {"min_value": min3d, "max_value": max3d, "total_value": sum3d, "avg_value": avg3d}
    return dat, stats


def make_plots(data=None, work_dir="", get_figure=False):
    """
    Creates png and svg plots of Aggrescan3D scores for a single chain
    :param data: dictionary - {chainID:[residue ID, agg3d score, residue label]}
    :param work_dir: directory where the plots will be saved
    :param get_figure: if set to True, will return the figure
    :return: None or plt.figure object
    """

    warnings.simplefilter("ignore")
    for chain in data.keys():
        if chain != "All":
            dat = data[chain]

            params()
            fig = plt.figure(figsize=(10, 6.6))
            x = np.array([l[0] for l in dat])
            y = np.array([l[1] for l in dat])
            l = np.array([l[2] for l in dat])

            plt.xlabel("Residue")
            plt.ylabel("Score")
            plt.axhline(linewidth=1, color='0.45', linestyle='--')
            plt.xticks(x[1::10], l[1::10], rotation=35, fontsize='small')
            plt.title("A3D profile | chain " + chain)
            plt.axis(ymin=-4, ymax=4, xmin=min(x) - 2, xmax=max(x) + 2)
            plt.plot(x, y, linewidth=1.5, alpha=0.75, marker='o', mec='None')
            plt.grid(alpha=0.5, color='0.9', linewidth=1, linestyle='--')

            for x, y, l in zip(x, y, l):
                if float(y) > 0.0:
                    plt.annotate(l, xy=(x, y), xytext=(1, 1), alpha=0.5, fontsize='small', gid="label_" + str(x),
                                 textcoords='offset points')
            logger.log_file(module_name=_name,msg="Saving plots as %s.png and %s.svg" % (chain, chain))
            plt.savefig(os.path.join(work_dir, "%s.png" % chain), format="png")
            plt.savefig(os.path.join(work_dir, "%s.svg" % chain), format="svg")
            if get_figure: return fig


def make_auto_mut_plot(work_dir=""):
    """
    Create a collective plot of mutants and the wild type, this is mostly a copy paste from server plot into mpl
    #TODO actually use fig axes object rather than plt like that
    """
    _target_mutations = ["E", "K", "D", "R"]    # This should mirror one in auto_mutation but can't be imported because they also import this first
    warnings.simplefilter("ignore")
    mutants = []
    with open(join(work_dir, "Mutations_summary.csv"), "r") as f:
        f.readline()
        for line in f:
            mutants.append(line.split(",")[0])

    with open(join(work_dir, "A3D.csv"), 'r') as f:
        f.readline()  # skip the initial line with labels
        wild_labels, wild_y = [], []
        for line in f:
            a = line.strip().split(',')
            # a goes as follows: model name, chain, index, one letter code, aggrescan score
            wild_labels.append(("Chain %s" % a[1], a[2] + a[3]))
            wild_y.append(float(a[-1]))
        wild_x = [i for i in range(len(wild_y))]
    while mutants:
        data = OrderedDict()
        one_r_mutated = []
        mutated = mutants[0][0]
        for mutant in mutants[:]:   # Create a slice to prevent the iterator from skipping itmes
            if mutant[0] == mutated:
                one_r_mutated.append(mutants.pop(mutants.index(mutant)))    # Mutations are 'guaranteed' to be unique
        data["Wild_type"] = [wild_x, wild_y, wild_labels]
        for mutant in one_r_mutated:
            with open(join(work_dir, mutant + ".csv"), 'r') as f:
                f.readline()    # skip the initial line with labels
                labels, y = [], []
                for line in f:
                    a = line.strip().split(',')
                    # a goes as follows: model name, chain, index, one letter code, aggrescan score
                    labels.append(("Chain %s" % a[1], a[2]+a[3]))
                    y.append(float(a[-1]))
            x = [i for i in range(len(y))]
            data[mutant] = [x, y, labels]
        _plot(data, work_dir, filename="%s_mutants" % mutant[2:])


def _plot(data, work_dir, filename):
    params()
    fig = plt.figure(figsize=(10, 6.6))
    plt.ylabel("Score")
    plt.axhline(linewidth=1, color='0.45', linestyle='--')
    plt.title("A3D mutations profile")
    plt.grid(alpha=0.5, color='0.9', linewidth=1, linestyle='--')
    for key, value in data.iteritems():
        x, y, labels = value
        plt.plot(x, y, label=key, linewidth=1.5, alpha=0.75, marker='o', mec='None')
    plt.xticks(x[1::10], labels[1::10], rotation=35, fontsize='small')
    logger.log_file(module_name=_name, msg="Saving auto mutation plots to %s (svg and png)" % filename)
    plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
    plt.savefig(os.path.join(work_dir, filename + ".png"), format="png",  bbox_inches='tight')
    plt.savefig(os.path.join(work_dir, filename + ".svg"), format="svg",  bbox_inches='tight')
    plt.close()


def append_scores(a3d="", in_pdb="", out_pdb="", work_dir=""):
    """
    Replaces the last field in pdb file with Aggrescan3D score
    :param a3d: filepath to a csv aggrescan-formatted file with scores
    :param in_pdb: filepath to a input pdb file
    :param out_pdb: fielapth to which the output will be written
    :param work_dir: Output directory
    :return: None
    """
    rec = re.compile(r"^(.*),(.*),(.*),(.*),(.*)$", re.M)
    amino_a_dict ={'A': 'ALA', 'R': 'ARG','N': 'ASN','D': 'ASP','C': 'CYS','E': 'GLU',
                   'Q': 'GLN','G': 'GLY','H': 'HIS','I': 'ILE','L': 'LEU','K': 'LYS',
                   'M': 'MET','F': 'PHE','P': 'PRO','S': 'SER','T':'THR','W': 'TRP',
                   'Y': 'TYR', 'V': 'VAL','X': 'UNK'}

    with open(a3d, "r") as agg_out_fh, open(in_pdb, "r") as p:
        block = p.read()
        d = rec.findall(agg_out_fh.read().replace("\r", ""))[1:]
        for r in d:
            amino_acid = amino_a_dict[r[3]]
            agg_score = "%6.2f" % (float(r[4]))
            res_details = "%3s %1s%4s" % (amino_acid, r[1], r[2])  # 3 letter code, chain ID, res ID
            block = re.sub(r'(?<=^ATOM.{13}'+res_details+'.{34})(.*)$', agg_score, block, flags=re.M)
    logger.to_file(filename=os.path.join(work_dir, out_pdb), content=block)


def save_stats(data="", work_dir="", output="statistics"):
    """
    Saves statistcs of the Aggrescan3D scores calculations
    :param data: string - JSON generated by get_scores, formatted like:
    {chainID:{"min_value":val,"max_value":val,"total_value":val,"avg_value":val}}
    :param work_dir: Output directory
    :param output: outputs filename
    :return: None
    """
    logger.to_file(filename=os.path.join(work_dir, output), content=data)


def prepare_output(work_dir="", final=True, model_name="", scores_to_pdb=False):
    """
    Calls make_plots, save_stats and append_scores, see details there
    :param work_dir: Working dir of Aggrescan3D run
    :param final: [bool] if True data is plotted and output.pdb generated to work_dir
    :param model_name: [string] filename of the currently analyzed pdb file (without the .pdb part)
    :param scores_to_pdb: [bool] Decide if a3d score should be pun in the file's bfactor place
    :return: dictionary - {chainID: {"min_value" : val, "max_value" : val,
                           "total_value" : val, "avg_value" : val}}
    """
    data, stats = get_scores(os.path.join(work_dir, "A3D.csv"))
    save_stats(data=json.dumps(stats), output=model_name + "_stats", work_dir=work_dir)
    if scores_to_pdb:
        append_scores(a3d=os.path.join(work_dir, "A3D.csv"), in_pdb=model_name + ".pdb",
                      out_pdb=model_name+".pdb", work_dir=work_dir)
    if final:
        make_plots(data=data, work_dir=work_dir)
        append_scores(a3d=os.path.join(work_dir, "A3D.csv"), in_pdb=os.path.join(work_dir, "folded.pdb"),
                      out_pdb=os.path.join(work_dir, "output.pdb"), work_dir=work_dir)
    return stats
